# -*- coding: utf-8 -*-
"""
plotly based plotting functions

@author: Jussi (jnu@iki.fi)
"""
from __future__ import division

from builtins import zip
import logging
from builtins import range
from collections import defaultdict
from functools import partial
import sys

import numpy as np
import plotly
import plotly.graph_objs as go
from plotly.matplotlylib.mpltools import merge_color_and_opacity
import plotly.tools
import plotly.subplots

from ..envutils import GaitDataError
from ..config import cfg
from ..stats import AvgTrial
from ..timedist import _pick_common_vars
from .. import models, normaldata, utils
from . import layouts
from .plot_common import (
    _get_cycle_name,
    _var_title,
    _cyclical_mapper,
    _style_mpl_to_plotly,
    _handle_cyclespec,
    _handle_style_and_color_args,
    _color_by_params,
    _style_by_params,
    _emg_yscale,
    _get_trial_cycles,
    _triage_var,
    _compose_varname,
    _nested_get,
)


logger = logging.getLogger(__name__)


def plot_extracted_box(curve_vals, vardefs):
    """Plot comparison of extracted gait curve values as box plot.

    Parameters
    ----------
    vardefs : list
        Nested list of variable definitions. The definitions are
    curve_vals : dict
        The curve extracted data, keyed by session.
    """

    contexts = utils.get_contexts()
    nvars = len(vardefs)
    subtitles = [_compose_varname(nested_keys) for nested_keys in vardefs]
    fig = plotly.subplots.make_subplots(rows=nvars, cols=1, subplot_titles=subtitles)
    legendgroups = set()

    # the plotting logic is a bit weird due to the way go.Box() works:
    # -we first consolidate one variable's data from all sessions into a 1-d array
    # -this is done separately for L/R context
    # -the consolidated data is then plotted along with the session identifiers
    for row, vardef in enumerate(vardefs):
        for ctxt, _ in contexts:
            vals = list()
            groupnames = list()
            vardef_ctxt = [ctxt + vardef[0]] + vardef[1:]
            for session, session_vals in curve_vals.items():
                if vardef_ctxt[0] not in session_vals:
                    logger.debug(
                        '%s was not collected for this session' % vardef_ctxt[0]
                    )
                    continue
                this_vals = _nested_get(session_vals, vardef_ctxt)
                vals.extend(this_vals)
                # do not add session label on x axis if we only have a single session
                label = session if len(curve_vals) > 1 else ''
                groupnames.extend([label] * len(this_vals))
            # show entry in legend only if it was not already shown
            show_legend = ctxt not in legendgroups
            legendgroups.add(ctxt)
            box = go.Box(
                x=groupnames,
                y=vals,
                boxpoints=False,
                name=ctxt,
                offsetgroup=ctxt,
                legendgroup=ctxt,
                showlegend=show_legend,
                opacity=0.5,
                # mode='lines+markers',
                marker_color=cfg.plot.context_colors[ctxt],
            )
            fig.append_trace(box, row=row + 1, col=1)
            ylabel = _plotly_var_ylabel(vardef_ctxt[0])
            xaxis, yaxis = _get_plotly_axis_labels(row, 0, ncols=1)
            fig['layout'][yaxis].update(
                title={
                    'font': {'size': cfg.plot_plotly.label_fontsize},
                    'text': ylabel,
                    'standoff': 0,
                }
            )
    # group together boxes of the different traces for each value of x
    fig.update_layout(boxmode='group')
    # set subplot title font size
    for anno in fig['layout']['annotations']:
        anno['font']['size'] = cfg.plot_plotly.subtitle_fontsize
    return fig


def time_dist_barchart(
    values,
    stddev=None,
    thickness=0.5,
    color=None,
    stddev_bars=True,
    plotvars=None,
    big_fonts=False,
    figtitle=None,
):

    """Multi-variable and multi-condition barchart plot.

    Used mostly for plotting time-distance variables.

    Parameters
    ----------
    values : dict
        Nested dict of values to plot. The data values are given in
        values[condition][var][context] where context is 'Left' or 'Right' and
        their units in values[condition][var]['unit']. Usually it is the output
        of timedist._group_analysis_trials().
    stddev : dict | None
        Dict with structure similar to values, but provides standard deviation
        for each variable.
    thickness : float, optional
        Y direction thickness of the bars, by default 0.5
    color : list, optional
        List of colors for the different conditions. If None, use default values.
    stddev_bars : bool, optional
        Whether to plot standard deviation also as bars.
    plotvars : list, optional
        The variables to plot and their order. If None, plot all variables.
    figtitle : str, optional
        Title of the plot.
    big_fonts : bool, optional
        If True, increase font sizes somewhat.

    Returns
    -------
    dict
        The chart.
    """

    conds, vars, units = _pick_common_vars(values, plotvars)
    vars = vars[::-1]  # plotly yaxis starts from bottom
    units = units[::-1]

    legend_fontsize = cfg.plot_plotly.legend_fontsize
    label_fontsize = cfg.plot_plotly.label_fontsize
    subtitle_fontsize = cfg.plot_plotly.subtitle_fontsize
    if big_fonts:
        legend_fontsize += 2
        label_fontsize += 2
        subtitle_fontsize += 2

    data = dict()
    texts = dict()
    ctxts = ['Left', 'Right']
    for cond in conds:
        data[cond] = dict()
        texts[cond] = dict()
        for ctxt in ctxts:
            # flatten data into simple arrays of nvars x 1
            data[cond][ctxt] = np.array([values[cond][var][ctxt] for var in vars])
            if stddev:
                stddevs = np.array([stddev[cond][var][ctxt] for var in vars])
            if stddev and stddevs.max() > 0:
                texts[cond][ctxt] = [
                    u'%.2f ± %.2f %s' % (val, std, unit)
                    for val, std, unit in zip(data[cond][ctxt], stddevs, units)
                ]
            else:
                texts[cond][ctxt] = [
                    u'%.2f %s' % (val, unit)
                    for val, unit in zip(data[cond][ctxt], units)
                ]

    # scale vars according to their maximums over all conditions
    scaler = dict()
    for ctxt in ctxts:
        scaler[ctxt] = np.max(np.array([data[c][ctxt] for c in conds]), axis=0)
    for cond in conds:
        for ctxt in ctxts:
            data[cond][ctxt] /= scaler[ctxt] * 0.01

    fig = plotly.subplots.make_subplots(
        rows=1,
        cols=2,
        specs=[[{}, {}]],
        shared_xaxes=True,
        shared_yaxes=True,
        vertical_spacing=0,
        horizontal_spacing=0.05,
        subplot_titles=ctxts,
    )

    # ordering the bars properly is a bit tricky. seemingly, there's no way to control
    # ordering of bars within a category, and they seem to be plotted from bottom to up by default.
    # a dirty solution is to plot in reversed order and then reverse the legend also.
    varlabels = [s + ' ' for s in vars]  # hack: add spaces to create some margin
    for condn, cond in enumerate(reversed(conds)):
        barcolor = cfg.plot.colors[condn]
        for k, ctxt in enumerate(ctxts, 1):
            show_legend = k == 1
            trace_l = go.Bar(
                y=varlabels,
                x=data[cond][ctxt],
                orientation='h',
                name=cond,
                legendgroup=cond,
                text=texts[cond][ctxt],
                textfont={'size': label_fontsize + 2},
                textposition='auto',
                showlegend=show_legend,
                hoverlabel=dict(namelength=-1),
                hoverinfo='y+text+name',
                marker_color=barcolor,
            )
            fig.add_trace(trace_l, 1, k)
            fig['layout']['legend']['traceorder'] = 'reversed'
            # increase var label size a bit
            fig['layout']['yaxis%d' % k].update(tickfont={'size': label_fontsize + 2})
            fig['layout']['xaxis%d' % k].update(
                title={'text': '% of maximum', 'font': {'size': label_fontsize}}
            )

    margin = go.layout.Margin(l=50, r=0, b=50, t=50, pad=4)  # NOQA: 741
    legend = dict(font=dict(size=legend_fontsize))
    plotly_layout = go.Layout(
        margin=margin,
        legend=legend,
        paper_bgcolor='rgba(255,255,255,0)',  # no background please
        plot_bgcolor='rgba(255,255,255,0)',
        font={'size': label_fontsize},
        hovermode='closest',
        title=figtitle,
    )

    fig['layout'].update(plotly_layout)
    for anno in fig['layout']['annotations']:
        anno['font']['size'] = subtitle_fontsize

    return fig


def _plot_vels(vels, labels, title=None):
    """Plot trial velocities as a stem plot"""
    trace = dict(y=vels, x=labels, mode='markers')
    layout = dict(
        title=title,
        xaxis=dict(title='Trial', automargin=True),
        yaxis=dict(title='Speed (m/s)'),
    )
    return dict(data=[trace], layout=layout)


def _plot_timedep_vels(vels, labels, title=None):
    """Plot trial time-dependent velocities"""
    traces = list()
    for vel, label in zip(vels, labels):
        trace = dict(y=vel, text=label, name=label, hoverinfo='x+y+text')
        traces.append(trace)
    # FIXME: labels get truncated, not fixed by automargin
    layout = dict(
        title=title,
        xaxis=dict(title='% of trial', automargin=True),
        yaxis=dict(title='Velocity (m/s)'),
    )
    return dict(data=traces, layout=layout)


def _plotly_fill_between(x, ylow, yhigh, **kwargs):
    """Fill area between ylow and yhigh"""
    x_ = np.concatenate([x, x[::-1]])  # construct a closed curve
    y_ = np.concatenate([yhigh, ylow[::-1]])
    return dict(x=x_, y=y_, fill='toself', mode='none', hoverinfo='none', **kwargs)


def plot_trials_browser(trials, layout, **kwargs):
    """Convenience plotter, uses plotly.offline to plot directly to browser"""
    fig = plot_trials(trials, layout, **kwargs)
    plotly.offline.plot(fig)


def _get_plotly_axis_labels(row, col, ncols):
    """Gets plotly axis labels from zero-based row and col indices"""
    plot_ind = row * ncols + col + 1  # plotly subplot index
    return 'xaxis%d' % plot_ind, 'yaxis%d' % plot_ind


def _plotly_var_ylabel(var, themodel=None):
    """Make ylabel for model variable var"""
    if themodel is None:
        themodel = models.model_from_var(var)
    yunit = themodel.units[var]
    if yunit == 'deg':
        yunit = u'\u00B0'  # Unicode degree sign
    ydesc = [s[:3] for s in themodel.ydesc[var]]  # shorten
    ylabel = u'%s %s %s' % (ydesc[0], yunit, ydesc[1])
    # Py2: plotly cannot handle unicode objects
    if sys.version_info.major == 2 and isinstance(ylabel, unicode):
        ylabel = ylabel.encode('utf-8')
    return ylabel


def plot_trials(
    trials,
    layout,
    model_normaldata=None,
    emg_normaldata=None,
    cycles=None,
    max_cycles=None,
    emg_mode=None,
    legend_type=None,
    style_by=None,
    color_by=None,
    supplementary_data=None,
    legend=True,
    figtitle=None,
    big_fonts=False,
):
    """Plot gait trials using Plotly.

    Parameters
    ----------
    trials : list
        List of Trial instances to plot.
    layout : list,
        The plot layout to use.
    model_normaldata : dict | None
        Normaldata for model variables. If None, taken from cfg.
    emg_normaldata : dict | None
        Normal data for EMG variables. If None, taken from cfg.
    cycles : dict | str | int | tuple | list
        Cycles to plot. See Trial.get_cycles() for details.
    max_cycles : dict | None
        Maximum number of cycles to plot for each variable type. If None, taken
        from cfg.
    emg_mode : str | None
        If 'rms', plot EMG in RMS mode.
    legend_type : str | None
        Legend type for gait cycles (see _get_cycle_name for options). If None,
        taken from cfg.
    style_by : dict | None
        How to style each variable type. If None, taken from cfg.
    color_by : dict | None
        How to color each variable type. If None, taken from cfg.
    supplementary_data : dict | None
        Supplementary data to plot for each variable.
    legend : bool
        If True, plot the legend.
    figtitle : str | None
        Main title for the figure.
    big_fonts : bool
        If True, increase font sizes somewhat.

    Returns
    -------
    dict
        The Plotly figure object.
    """

    if not trials:
        raise GaitDataError('No trials')

    if not isinstance(trials, list):
        trials = [trials]

    style_by, color_by = _handle_style_and_color_args(style_by, color_by)

    if legend_type is None:
        legend_type = 'short_name_with_cyclename'

    if supplementary_data is None:
        supplementary_data = dict()

    if model_normaldata is None:
        model_normaldata = normaldata._read_configured_model_normaldata()

    if emg_normaldata is None:
        emg_normaldata = normaldata._read_configured_emg_normaldata()

    use_rms = emg_mode == 'rms'

    nrows, ncols = layouts._check_layout(layout)

    # these generate and keep track of key -> linestyle (or color) mappings
    trace_colors = _cyclical_mapper(cfg.plot.colors)
    emg_trace_colors = _cyclical_mapper(cfg.plot.colors)
    trace_styles = _cyclical_mapper(cfg.plot.linestyles)

    allvars = [item for row in layout for item in row]
    titles = [_var_title(var) for var in allvars]
    fig = plotly.subplots.make_subplots(
        rows=nrows, cols=ncols, print_grid=False, subplot_titles=titles
    )
    legendgroups = set()
    model_normaldata_legend = True
    emg_normaldata_legend = True

    normalized = cycles != 'unnormalized'
    cycles = _handle_cyclespec(cycles)
    if max_cycles is None:
        max_cycles = cfg.plot.max_cycles

    legend_fontsize = cfg.plot_plotly.legend_fontsize
    label_fontsize = cfg.plot_plotly.label_fontsize
    subtitle_fontsize = cfg.plot_plotly.subtitle_fontsize
    if big_fonts:
        legend_fontsize += 2
        label_fontsize += 2
        subtitle_fontsize += 2

    # plot normaldata first to ensure that its z order is lowest
    # and it gets the 1st legend entries
    if normalized:
        for i, row in enumerate(layout):
            for j, var in enumerate(row):
                themodel = models.model_from_var(var)
                if themodel and model_normaldata:
                    nvar = var if var in themodel.varlabels_noside else var[1:]
                    key = nvar if nvar in model_normaldata else None
                    ndata = model_normaldata[key] if key in model_normaldata else None
                    if ndata is not None:
                        normalx = np.linspace(0, 100, ndata.shape[0])
                        fillcolor = merge_color_and_opacity(
                            cfg.plot.model_normals_color, cfg.plot.model_normals_alpha
                        )
                        ntrace = _plotly_fill_between(
                            normalx,
                            ndata[:, 0],
                            ndata[:, 1],
                            fillcolor=fillcolor,
                            name='Norm.',
                            legendgroup='Norm.',
                            showlegend=model_normaldata_legend,
                            line=dict(width=0),
                        )  # no border lines
                        fig.add_trace(ntrace, i + 1, j + 1)
                        model_normaldata_legend = False  # mark as plotted

                elif var in cfg.emg.channel_labels and var in emg_normaldata:
                    # build x, y, z triplets for heatmap
                    # cell size is automatically determined from y values, which is a bit clumsy
                    # the idea is to build two strips of normal data at nearby y values, which fixes
                    # the cell size at a small value (dy)
                    _emg_y_extent = _emg_yscale(emg_mode)
                    extent_y0 = _emg_y_extent[0]
                    extent_y1 = extent_y0 + (_emg_y_extent[1] - _emg_y_extent[0]) / 20.0
                    Npts = 101
                    x = np.concatenate(
                        (np.linspace(0, 100, Npts), np.linspace(0, 100, Npts))
                    )
                    y = np.concatenate(
                        (extent_y0 * np.ones(Npts), extent_y1 * np.ones(Npts))
                    )
                    z = np.concatenate((emg_normaldata[var], emg_normaldata[var]))
                    heatmap = go.Heatmap(
                        z=z,
                        y=y,
                        x=x,
                        colorscale='reds',
                        zmin=0,
                        zmax=1,
                        opacity=0.5,
                        showscale=False,
                        name='EMG norm.',
                        legendgroup='EMG norm.',
                        showlegend=emg_normaldata_legend,
                    )
                    fig.add_trace(heatmap, i + 1, j + 1)
                    emg_normaldata_legend = False  # mark as plotted

    # plot the actual data
    for trial in trials:
        subplot_adjusted = defaultdict(lambda: False)
        cyclebunch = _get_trial_cycles(trial, cycles, max_cycles)
        # the idea here is to sort the trial cycles by their legend key, so they
        # appear in the legend in correct order
        sorter = partial(_get_cycle_name, trial, name_type=legend_type)
        allcycles = sorted(cyclebunch.allcycles, key=sorter)

        for cyc in allcycles:

            context = cyc.context

            for i, row in enumerate(layout):
                for j, var in enumerate(row):

                    vartype = _triage_var(var, trial)
                    if vartype is None:
                        continue

                    xaxis, yaxis = _get_plotly_axis_labels(i, j, ncols)

                    cyclename = _get_cycle_name(trial, cyc, name_type=legend_type)
                    cyclename_full = _get_cycle_name(trial, cyc, name_type='full')

                    # Py2: plotly cannot handle unicode objects
                    if sys.version_info.major == 2 and isinstance(cyclename, unicode):
                        cyclename = cyclename.encode('utf-8')

                    if vartype == 'model':
                        do_plot = cyc in cyclebunch.model_cycles
                        themodel = models.model_from_var(var)
                        if var in themodel.varnames_noside:
                            # var context was unspecified, so choose it
                            # according to cycle context
                            var = context + var
                        elif var[0] != context:
                            # var context was specified and does not match cycle
                            do_plot = False

                        # kinetic var cycles are required to have valid
                        # forceplate data
                        if (
                            normalized
                            and themodel.is_kinetic_var(var)
                            and not cyc.on_forceplate
                        ):
                            do_plot = False

                        t, y = trial.get_model_data(var, cycle=cyc)
                        if y is None:
                            do_plot = False

                        if do_plot:
                            sty = _style_by_params(
                                style_by['model'], trace_styles, trial, cyc, context
                            )
                            sty = _style_mpl_to_plotly(sty)
                            col = _color_by_params(
                                color_by['model'], trace_colors, trial, cyc, context
                            )

                            line = dict(
                                width=cfg.plot.model_linewidth, dash=sty, color=col
                            )

                            # for model variables, put traces into legend groups
                            # according to trial/cycle info
                            legendgroup = cyclename
                            show_legend = cyclename not in legendgroups

                            trace = dict(
                                x=t,
                                y=y,
                                name=cyclename,
                                text=cyclename_full,
                                legendgroup=legendgroup,
                                showlegend=show_legend,
                                hoverlabel=dict(namelength=-1),
                                hoverinfo='x+y+text',
                                line=line,
                            )

                            # add toeoff marker
                            if cyc.toeoffn is not None:
                                marker = dict(color=col, symbol='triangle-up', size=8)
                                toeoff_marker = dict(
                                    x=t[cyc.toeoffn : cyc.toeoffn + 1],
                                    y=y[cyc.toeoffn : cyc.toeoffn + 1],
                                    showlegend=False,
                                    legendgroup=legendgroup,
                                    hoverinfo='skip',
                                    mode='markers',
                                    marker=marker,
                                )
                                fig.add_trace(toeoff_marker, i + 1, j + 1)

                            # add trace to figure
                            fig.add_trace(trace, i + 1, j + 1)
                            legendgroups.add(legendgroup)

                            # each cycle gets its own stddev plot
                            if isinstance(trial, AvgTrial):
                                model_stddev = trial.stddev_data
                                if (
                                    model_stddev is not None
                                    and normalized
                                    and y is not None
                                    and var in model_stddev
                                ):
                                    sdata = model_stddev[var]
                                    stdx = np.linspace(0, 100, sdata.shape[0])
                                    fillcolor = merge_color_and_opacity(
                                        col,
                                        cfg.plot.model_stddev_alpha,
                                    )
                                    ntrace = _plotly_fill_between(
                                        stdx,
                                        y - sdata,
                                        y + sdata,
                                        fillcolor=fillcolor,
                                        name='Stddev, %s' % cyclename,
                                        legendgroup='Stddev, %s' % cyclename,
                                        showlegend=show_legend,
                                        line=dict(width=0),
                                    )  # no border lines
                                    fig.add_trace(ntrace, i + 1, j + 1)

                            # add supplementary data
                            if cyc in supplementary_data:
                                supdata = supplementary_data[cyc]
                                if var in supdata:
                                    logger.debug(
                                        'plotting supplementary data '
                                        'for var %s' % var
                                    )
                                    t_sup = supdata[var]['t']
                                    data_sup = supdata[var]['data']
                                    label_sup = supdata[var]['label']

                                    strace = dict(
                                        x=t_sup,
                                        y=data_sup,
                                        name=label_sup,
                                        text=label_sup,
                                        line=line,
                                        legendgroup=cyclename,
                                        hoverinfo='x+y+text',
                                        showlegend=False,
                                    )
                                    fig.add_trace(strace, i + 1, j + 1)
                                    legendgroups.add(cyclename)

                            # adjust subplot once
                            if not subplot_adjusted[(i, j)]:
                                ylabel = _plotly_var_ylabel(var, themodel)
                                fig['layout'][yaxis].update(
                                    title={
                                        'text': ylabel,
                                        'font': {'size': label_fontsize},
                                        'standoff': 0,
                                    }
                                )
                                # less decimals on hover label
                                fig['layout'][yaxis].update(hoverformat='.2f')
                                subplot_adjusted[(i, j)] = True

                    elif vartype == 'marker':
                        do_plot = cyc in cyclebunch.marker_cycles
                        t, mdata = trial.get_marker_data(var, cycle=cyc)
                        if mdata is None:
                            do_plot = False

                        if do_plot:

                            for datadim, data in zip('XYZ', mdata.T):
                                sty = _style_by_params(
                                    style_by['marker'],
                                    trace_styles,
                                    trial,
                                    cyc,
                                    context,
                                    datadim,
                                )
                                sty = _style_mpl_to_plotly(sty)
                                col = _color_by_params(
                                    color_by['marker'],
                                    trace_colors,
                                    trial,
                                    cyc,
                                    context,
                                    datadim,
                                )
                                line = dict(
                                    width=cfg.plot.model_linewidth, dash=sty, color=col
                                )
                                # dim-specific tracename
                                tracename_marker = 'mkr_%s:%s' % (datadim, cyclename)
                                # dimension- and trial-based grouping
                                # legendgroup = 'mkr_%s:%s' % (datadim, trial.trialname)
                                # dimension- and cycle-based grouping
                                legendgroup = tracename_marker
                                show_legend = legendgroup not in legendgroups
                                trace = dict(
                                    x=t,
                                    y=data,
                                    name=tracename_marker,
                                    text=cyclename_full,
                                    legendgroup=legendgroup,
                                    showlegend=show_legend,
                                    hoverlabel=dict(namelength=-1),
                                    hoverinfo='x+y+name',
                                    line=line,
                                )
                                fig.add_trace(trace, i + 1, j + 1)
                                legendgroups.add(legendgroup)

                                # add toeoff marker
                                if cyc.toeoffn is not None:
                                    cyc.toeoffn = int(cyc.toeoffn)
                                    marker = dict(
                                        color=col, symbol='triangle-up', size=8
                                    )
                                    toeoff_marker = dict(
                                        x=t[cyc.toeoffn : cyc.toeoffn + 1],
                                        y=data[cyc.toeoffn : cyc.toeoffn + 1],
                                        showlegend=False,
                                        legendgroup=legendgroup,
                                        hoverinfo='skip',
                                        mode='markers',
                                        marker=marker,
                                    )
                                    fig.add_trace(toeoff_marker, i + 1, j + 1)

                            # adjust subplot once
                            if not subplot_adjusted[(i, j)]:
                                # fig['layout'][xaxis].update(showticklabels=False)
                                ylabel = 'mm'
                                # Py2: plotly cannot handle unicode objects
                                if sys.version_info.major == 2 and isinstance(
                                    ylabel, unicode
                                ):
                                    ylabel = ylabel.encode('utf-8')
                                fig['layout'][yaxis].update(
                                    title={
                                        'text': ylabel,
                                        'font': {'size': label_fontsize},
                                        'standoff': 0,
                                    }
                                )
                                # less decimals on hover label
                                fig['layout'][yaxis].update(hoverformat='.2f')
                                subplot_adjusted[(i, j)] = True

                    # plot EMG variable
                    elif vartype == 'emg':
                        do_plot = (
                            trial.emg.context_ok(var, cyc.context)
                            and trial.emg.status_ok(var)
                            and cyc in cyclebunch.emg_cycles
                        )
                        # FIXME: maybe annotate disconnected chans
                        # _no_ticks_or_labels(ax)
                        # _axis_annotate(ax, 'disconnected')
                        if do_plot:
                            tracename_emg = 'EMG:' + cyclename

                            t_, y = trial.get_emg_data(var, rms=use_rms, cycle=cyc)
                            t = t_ if normalized else t_ / trial.samplesperframe

                            col = _color_by_params(
                                color_by['emg'], emg_trace_colors, trial, cyc, context
                            )
                            col = merge_color_and_opacity(col, cfg.plot.emg_alpha)
                            lw = (
                                cfg.plot.emg_rms_linewidth
                                if use_rms
                                else cfg.plot.emg_linewidth
                            )
                            line = {'width': lw, 'color': col}

                            # EMG traces get grouped according to cycle (same
                            # legendgroup as model traces)
                            # the tracename_emg legend group does not actually exist
                            # in plotly, it's only used to keep track of whether the
                            # EMG trace legend was already shown.
                            show_legend = tracename_emg not in legendgroups
                            legendgroup = cyclename

                            trace = dict(
                                x=t,
                                y=y * cfg.plot.emg_multiplier,
                                name=tracename_emg,
                                legendgroup=legendgroup,
                                showlegend=show_legend,
                                line=line,
                            )
                            legendgroups.add(tracename_emg)
                            fig.add_trace(trace, i + 1, j + 1)

                        # adjust subplot once
                        if not subplot_adjusted[(i, j)]:
                            fig['layout'][yaxis].update(
                                title={
                                    'text': cfg.plot.emg_ylabel,
                                    'font': {'size': label_fontsize},
                                    'standoff': 0,
                                },
                                range=_emg_yscale(emg_mode),
                            )
                            # prevent changes due to legend clicks etc.
                            if normalized:
                                fig['layout'][xaxis].update(range=[0, 100])
                            # rm x tick labels, plot too crowded
                            # fig['layout'][xaxis].update(showticklabels=False)
                            # less decimals on hover label
                            # XXX: this is not working?
                            fig['layout'][yaxis].update(hoverformat='.2f')
                            subplot_adjusted[(i, j)] = True

                    elif vartype == 'unknown':
                        raise GaitDataError('cannot interpret variable %s' % var)

                    else:
                        raise GaitDataError(
                            'plotting not implemented for variable %s' % var
                        )

    # set subplot title font size
    for anno in fig['layout']['annotations']:
        anno['font']['size'] = subtitle_fontsize

    # put x labels on last row only, re-enable tick labels for last row
    inds_last = range((nrows - 1) * ncols, nrows * ncols)
    axes_last = ['xaxis%d' % (ind + 1) for ind in inds_last]
    xlabel = '% of gait cycle' if normalized else 'frame'
    for ax in axes_last:
        fig['layout'][ax].update(
            title={'text': xlabel, 'font': {'size': label_fontsize}},
            showticklabels=True,
        )

    margin = go.layout.Margin(l=50, r=0, b=50, t=50, pad=4)  # NOQA: 741
    legend = dict(font=dict(size=legend_fontsize))
    plotly_layout = go.Layout(
        margin=margin,
        legend=legend,
        font={'size': label_fontsize},
        hovermode='closest',
        title=figtitle,
    )

    fig['layout'].update(plotly_layout)
    return fig



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Time-distance plots.

@author: Jussi (jnu@iki.fi)
"""

import logging
import os.path as op
from collections import OrderedDict

from ..envutils import GaitDataError
from ..config import cfg
from .. import sessionutils
from ..timedist import _group_analysis_trials
from ..report.text import _timedist_vars
from .plot_misc import get_backend

logger = logging.getLogger(__name__)


def plot_session_average(session, tags=None, backend=None):
    """Plot time-distance average of session trials.

    Parameters
    ----------
    session : str
        The session directory.
    tags : list | None, optional
        Pick trials with given Eclipse tags. None to use default tags from
        config. To collect all trials, use tags=[].
    backend : str | None
        Name of backend to use, currently 'plotly' or 'matplotlib'. If None,
        taken from cfg.

    Returns
    -------
    fig : Figure | dict
        The figure object. Type depends on backend. Use show_fig() to show it.
    """
    if tags is None:
        tags = cfg.eclipse.tags
    trials = sessionutils.get_c3ds(session, tags=tags, trial_type='dynamic')
    if not trials:
        raise GaitDataError('No tagged trials found for session %s' % session)
    session_ = op.split(session)[-1]
    fig = plot_trials(
        {session_: trials},
        title='Time-distance average, session %s' % session_,
        backend=backend,
    )
    return fig


def plot_comparison(sessions, tags=None, big_fonts=False, backend=None):
    """Plot time-dist comparison of multiple sessions.

    Parameters
    ----------
    sessions : list
        The session directories.
    tags : list | None, optional
        Pick trials with given Eclipse tags. None to use default tags from
        config. To collect all trials, use tags=[].
    big_fonts : bool, optional
        Increase font size (plotly backend only).
    backend : str | None
        Name of backend to use, currently 'plotly' or 'matplotlib'. If None,
        taken from cfg.

    Returns
    -------
    fig : Figure | dict
        The figure object. Type depends on backend. Use show_fig() to show it.
    """
    if tags is None:
        tags = cfg.eclipse.tags
    trials = OrderedDict()

    for session in sessions:
        c3ds = sessionutils.get_c3ds(session, tags=tags, trial_type='dynamic')
        if not c3ds:
            raise RuntimeError('No tagged trials found in session %s' % session)
        cond_label = op.split(session)[-1]
        trials[cond_label] = c3ds

    return plot_trials(trials, big_fonts=big_fonts, backend=backend)


def plot_trials(c3dfiles, plotvars=None, title=None, big_fonts=False, backend=None):
    """Plot a time-distance barchart from given c3d files.

    Parameters
    ----------
    c3dfiles : dict
        The c3d files to plot. Dict of lists keyed by condition name.
    plotvars : list, optional
        The time-distance variables to plot. This also determines the plotting
        order.
    title : str, optional
        Plot title.
    big_fonts : bool, optional
        Increase font size (plotly backend only).
    backend : str | None
        Name of backend to use, currently 'plotly' or 'matplotlib'. If None,
        taken from cfg.

    Returns
    -------
    fig : Figure | dict
        The figure object. Type depends on backend. Use show_fig() to show it.
    """
    if plotvars is None:
        plotvars = _timedist_vars
    res_avg_all, res_std_all = _group_analysis_trials(c3dfiles)
    backend_lib = get_backend(backend)
    return backend_lib.time_dist_barchart(
        res_avg_all,
        stddev=res_std_all,
        stddev_bars=False,
        plotvars=plotvars,
        figtitle=title,
        big_fonts=big_fonts,
    )


def _plot_single_trial(c3dfile, backend=None):
    """Plot single trial time-distance"""
    _, c3dfile_ = op.split(c3dfile)
    fig = plot_trials(
        {c3dfile_: [c3dfile]},
        title='Time-distance variables, %s' % c3dfile_,
        backend=backend,
    )
    return fig


def _plot_multitrial(c3dfiles, show=True, backend=None):
    """Plot multiple trial comparison time-distance"""
    trials = {op.split(c3d)[-1]: [c3d] for c3d in c3dfiles}
    return plot_trials(trials, backend=backend)



# -*- coding: utf-8 -*-
"""
@author: Jussi (jnu@iki.fi)
"""

from setuptools import setup, find_packages

# entry points for console scripts
console_entries = [
    'gaitmenu=gaitutils.gui._gaitmenu:main',
    'tardieu=gaitutils.gui._tardieu:main',
    'nexus_autoproc_session=gaitutils.autoprocess:autoproc_session',
    'nexus_autoproc_trial=gaitutils.autoprocess:autoproc_trial',
    'nexus_automark_trial=gaitutils.autoprocess:automark_trial',
    'nexus_plot_session_avg=gaitutils.viz.console:plot_nexus_session_average',
    'nexus_plot_session=gaitutils.viz.console:plot_nexus_session',
    'gaitmenu_make_shortcut=gaitutils.envutils:_make_gaitutils_shortcut',
]

setup(
    name='gaitutils',
    version='0.13.6',
    description='Utilities for processing and plotting gait data',
    author='Jussi Nurminen',
    author_email='jnu@iki.fi',
    license='GPLv3',
    url='https://github.com/jjnurminen/gaitutils',
    packages=find_packages(),
    entry_points={'console_scripts': console_entries},
    include_package_data=True,
)



# -*- coding: utf-8 -*-
"""
Created on Tue Mar 17 14:41:31 2015

Read gait trials.

@author: Jussi (jnu@iki.fi)
"""


from __future__ import division

from builtins import range
from builtins import object
from collections import defaultdict
import numpy as np
import re
import os.path as op
import logging

from .emg import EMG
from .config import cfg
from .envutils import GaitDataError
from . import (
    read_data,
    nexus,
    utils,
    eclipse,
    models,
    videos,
    sessionutils,
)

logger = logging.getLogger(__name__)


def nexus_trial(from_c3d=False):
    """Return Trial instance created from the currently open Nexus trial.

    Parameters
    ----------
    from_c3d : bool
        If True, try to read the trial data from the corresponding c3d file.
        If False, read via Nexus API (slower).

    Returns
    -------
    Trial
        A Trial instance.
    """
    vicon = nexus.viconnexus()
    trname = vicon.GetTrialName()  # 2-tuple of (path, name)
    if not trname[1]:
        raise GaitDataError('No trial loaded in Nexus')
    if from_c3d:
        c3dfile = op.join(*trname) + '.c3d'
        if op.isfile(c3dfile):
            return Trial(c3dfile)
        else:
            logger.info(
                'no c3d file %s for currently loaded trial, loading '
                'directly from Nexus' % c3dfile
            )
            return Trial(nexus.viconnexus())
    else:
        return Trial(nexus.viconnexus())


class Noncycle(object):
    """Used in place of Gaitcycle when requesting unnormalized data.

    Has a context parameter, to facilitate plotting unnormalized data for left
    and right separately.

    Parameters
    ----------
    context : str
        Cycle context: 'R' or 'L' for right and left, respectively.
    trial : Trial
        The trial instance owning this cycle. Does not need to be set.
    """

    def __init__(self, context, trial=None):
        self.context = context
        self.trial = trial
        self.name = 'unnorm. (%s)' % context
        self.toeoffn = None
        self.on_forceplate = False
        self.start = 0  # to allow cycle sorting
        self.end = None


class Gaitcycle(object):
    """Gait cycle class.

    Parameters
    ----------
    start : int
        Starting frame for the cycle.
    start : int
        Ending frame for the cycle.
    toeoff : int
        Frame where toeoff occurs.
    context : str
        Cycle context: R or L for right and left, respectively.
    on_forceplate : bool
        Whether cycle starts on forceplate contact.
    plate_idx : int
        Index of forceplate.
    smp_per_frame : float
        Analog samples per frame.
    trial : instance of Trial
        The trial instance owning this cycle. Does not need to be set.
    name : str
        Name for the cycle. Can be set freely.
    index : int
        Cycle index.
    """

    def __init__(
        self,
        start,
        end,
        toeoff,
        context,
        on_forceplate,
        plate_idx,
        smp_per_frame,
        trial=None,
        name=None,
        index=None,
    ):
        self.len = end - start
        # convert frame indices to 0-based
        self.start = start
        self.end = end
        self.toeoff = toeoff
        # which foot begins and ends the cycle
        self.context = context
        # whether cycle begins with forceplate strike
        self.on_forceplate = on_forceplate
        # start and end on the analog samples axis; round to whole samples
        self.start_smp = int(round(self.start * smp_per_frame))
        self.end_smp = int(round(self.end * smp_per_frame))
        self.len_smp = self.end_smp - self.start_smp
        # normalized x-axis (% of gait cycle) of same length as cycle
        self.t = np.linspace(0, 100, self.len)
        # same for analog variables
        self.tn_analog = np.linspace(0, 100, self.len_smp)
        # normalized x-axis of 0,1,2..100%
        self.tn = np.linspace(0, 100, 101)
        # normalize toe-off event to the cycle
        self.toeoffn = int(round(100 * ((self.toeoff - self.start) / self.len)))
        self.trial = trial
        self.plate_idx = plate_idx
        self.index = index
        self.name = name

    def __repr__(self):
        s = '<Gaitcycle |'
        s += ' start: %d,' % self.start
        s += ' end: %d,' % self.end
        s += ' context: %s,' % self.context
        s += ' on forceplate,' if self.on_forceplate else ' not on forceplate,'
        s += ' toeoff: %d' % self.toeoff
        s += '>'
        return s

    def normalize(self, var):
        """Normalize (columns of) frames-based variable var to the cycle.

        Parameters
        ----------
        var : ndarray
            NxM array of frame-based data to normalize. N is the number of frames.

        Returns
        -------
        tuple
            A tuple of (tn, ndata) where tn is the normalized time (0..100%) and ndata
            is the normalized data.
        """
        # convert 1D arrays to 2D
        if len(var.shape) == 1:
            var = var[:, np.newaxis]
        ncols = var.shape[1]
        idata = np.array(
            [
                np.interp(self.tn, self.t, var[self.start : self.end, k])
                for k in range(ncols)
            ]
        ).T
        return self.tn, np.squeeze(idata)

    def crop_analog(self, var):
        """Crop analog variable (EMG, forceplate, etc. ) to the gait cycle.

        Parameters
        ----------
        var : ndarray
            NxM array of analog data to normalize.
        """
        return self.tn_analog, var[self.start_smp : self.end_smp]


class Trial(object):
    """Gait trial class.

    Parameters
    ----------
    source : str | instance of ViconNexus
        Source to read data from. Can be a c3d filename or a ViconNexus
        connection.

    Attributes
    ----------
    trialname : str
        Name of trial.
    eclipse_data : dict
        The Eclipse data for the trial. Keys are Eclipse fields and values are
        the corresponding data.
    sessionpath : str
        Full path to session directory.
    length : int
        Trial length in frames.
    offset : int
        Frame offset of data from beginning of trial. The event numbers will be
        interpreted relative to the offset, e.g. if the trial offset is 100 and
        an event occurs at frame 150, its index in the frame data array is 50.
    framerate : float
        Frame rate for capture (frames / sec).
    analograte : float
        Sampling rate for analog devices (samples / sec)-
    name : str
        Subject name.
    subj_params : dict
        Other subject parameters (bodymass etc.)
    events : TrialEvents
        Trial events (foot strikes, toeoffs etc.). These events are read from
        the trial data (i.e. not autodetected).
    """

    def __repr__(self):
        s = '<Trial |'
        s += ' trial: %s' % self.trialname
        s += ', data source: %s' % self.source
        s += ', subject: %s' % self.name
        s += ', gait cycles: %s' % self.ncycles
        s += '>'
        return s

    def __init__(self, source):
        logger.debug('new trial instance from %s' % source)
        self.source = source
        meta = read_data.get_metadata(source)
        # insert metadata dict directly as instance attributes (those are
        # documented above)
        self.__dict__.update(meta)
        # match events with frame data
        self.events.subtract_offset(self.offset)
        self.sessiondir = op.split(self.sessionpath)[-1]
        # try to locate trial .enf (we do not require it)
        enfpath = op.join(self.sessionpath, '%s.Trial.enf' % self.trialname)
        # also look for alternative (older style?) enf name
        if not op.isfile(enfpath):
            trialn_re = re.search('\.*(\d*)$', self.trialname)
            trialn = trialn_re.group(1)
            if trialn:
                trialname_ = '%s.Trial%s.enf' % (self.trialname, trialn)
                enfpath = op.join(self.sessionpath, trialname_)
        self.enfpath = enfpath
        if op.isfile(self.enfpath):
            logger.debug('reading Eclipse info from %s' % self.enfpath)
            edata = eclipse.get_eclipse_keys(self.enfpath)
            # for convenience, eclipse_data returns '' for nonexistent keys
            self.eclipse_data = defaultdict(lambda: '', edata)
        else:
            logger.debug('no .enf file found')
            self.eclipse_data = defaultdict(lambda: '', {})
        self.is_static = self.eclipse_data['TYPE'].upper() == 'STATIC'
        # handle session quirks
        quirks = sessionutils.load_quirks(self.sessionpath)
        if 'emg_correction_factor' in quirks:
            emg_correction_factor = quirks['emg_correction_factor']
            logger.warning(
                'using quirk: EMG correction factor = %g' % emg_correction_factor
            )
        else:
            emg_correction_factor = 1
        if 'ignore_eclipse_fp_info' in quirks:
            logger.warning('using quirk: ignore Eclipse forceplate fields')
            self.use_eclipse_fp_info = False
        else:
            self.use_eclipse_fp_info = True
        # data are lazily read
        self.emg = EMG(self.source, correction_factor=emg_correction_factor)
        self._forceplate_data = None
        self._marker_data = None
        if not self.is_static:
            self.fp_events = self._get_fp_events()
        else:
            self.fp_events = utils._empty_fp_events()
        self._models_data = dict()
        self.stddev_data = None  # AvgTrial only
        # frames 0...length
        self.t = np.arange(self.length)
        # analog frames 0...length
        self.t_analog = np.arange(self.length * self.samplesperframe)
        # normalized x-axis of 0, 1, 2 .. 100%
        self.tn = np.linspace(0, 100, 101)
        self.samplesperframe = self.analograte / self.framerate
        self.cycles = list() if self.is_static else list(self._scan_cycles())
        self.ncycles = len(self.cycles)

    @property
    def videos(self):
        """Returns names of trial video files.

        Returns
        -------
        vidfiles : list
            List of filenames.
        """
        trialbase = op.join(self.sessionpath, self.trialname)
        return videos.get_trial_videos(trialbase)

    @property
    def eclipse_tag(self):
        """Return the first matching Eclipse tag for the trial.

        The configured Eclipse fields (e.g. DESCRIPTION) are searched for the configured
        tags. If a configured tag is found, the it will be returned.

        Returns
        -------
        tag : str
            The tag.
        """
        for tag in cfg.eclipse.tags:
            if any([tag in self.eclipse_data[fld] for fld in cfg.eclipse.tag_keys]):
                return tag
        return None

    @property
    def name_with_description(self):
        """Return the trial name with Eclipse DESCRIPTION and NOTES keys.

        Returns
        -------
        desc : str
            The trial name and description.
        """
        return '%s (%s, %s)' % (
            self.trialname,
            self.eclipse_data['DESCRIPTION'],
            self.eclipse_data['NOTES'],
        )

    def normalize_to_cycle(self, data, cycle):
        """Normalize frame-based data to a gait cycle.

        Parameters
        ----------
        data : ndarray
            A Nxd ndarray, where N is number of frames and d is number of variables.
        cycle : Gaitcycle | Noncycle | None | int
            The gait cycle to normalize to. If Noncycle or None, returns unnormalized
            data. If int, return nth cycle from the trial cycles list.
        """
        if isinstance(cycle, int):
            if cycle >= len(self.cycles) or cycle < 0:
                raise ValueError('No such cycle')
            cycle = self.cycles[cycle]
        if isinstance(cycle, Gaitcycle):
            t, data = cycle.normalize(data)
        elif cycle is None or isinstance(cycle, Noncycle):
            # return unnormalized time axis and data as it is
            t = self.t
        else:
            raise ValueError('invalid type for cycle argument')
        return t, data

    def normalize_analog_to_cycle(self, data, cycle):
        """Normalize analog data to a gait cycle.

        Parameters
        ----------
        data : (Nxd) ndarray
            The data, where N is number of samples and d is number of variables.
        cycle : Gaitcycle | Noncycle | None
            The gait cycle to normalize to. If Noncycle or None, returns unnormalized
            data. If int, return nth cycle from the trial cycles list.
        """
        if isinstance(cycle, int):
            if cycle >= len(self.cycles) or cycle < 0:
                raise ValueError('No such cycle')
            cycle = self.cycles[cycle]
        if isinstance(cycle, Gaitcycle):
            t, data = cycle.crop_analog(data)
        elif cycle is None or isinstance(cycle, Noncycle):
            # return unnormalized time axis and data as it is
            t = self.t_analog
        else:
            raise ValueError('invalid type for cycle argument')
        return t, data

    @property
    def _full_marker_data(self):
        """Return the full marker data dict."""
        if self._marker_data is None:
            self._marker_data = read_data.get_marker_data(self.source, self.markers)
        return self._marker_data

    def _get_modelvar(self, var):
        """Return (unnormalized) data for a model variable."""
        model_ = models.model_from_var(var)
        if not model_:
            raise ValueError('No model found for %s' % var)
        if model_.desc not in self._models_data:
            # read and cache model data
            modeldata = read_data.get_model_data(self.source, model_)
            self._models_data[model_.desc] = modeldata
        return self._models_data[model_.desc][var]

    def get_model_data(self, var, cycle=None):
        """Return trial data for a model variable.

        Parameters
        ----------
        var : string
            The name of the model variable (e.g. 'LHipMomentX')
        cycle : Gaitcycle
            The cycle to normalize to. None for no normalization.

        Returns
        -------
        t_data : tuple
            Tuple of (t, data) where t is the time axis as 1-dim ndarray, and data
            is the model variable data as 1-dim ndarray.
        """
        data = self._get_modelvar(var)
        return self.normalize_to_cycle(data, cycle)

    def get_emg_data(self, ch, rms=False, cycle=None):
        """Return trial data for an EMG channel.

        Uses 'fuzzy' name matching: if the specified channel is not found in the
        data, partial name matches are considered and data for the shortest
        match is returned. For example, if ch == 'LGas' and the data has
        channels 'Voltage.LGas8' and 'Voltage.LGas8_dummy', the former is
        returned.

        Parameters
        ----------
        ch : string
            The EMG channel name. Fuzzy name matching is used.
        rms : bool
            Return moving-window RMS instead of raw data.

        Returns
        -------
        t_data : tuple
            Tuple of (t, data) where t is the time axis as 1-dim ndarray, and
            data is the EMG data as 1-dim ndarray.
        """
        data = self.emg.get_channel_data(ch, rms=rms)
        return self.normalize_analog_to_cycle(data, cycle)

    def get_marker_data(self, marker, cycle=None):
        """Return position data for a given marker.

        Parameters
        ----------
        marker : string
            The marker name.

        Returns
        -------
        t_data : tuple
            Tuple of (t, data) where t is the time axis as (Nt,) -shape ndarray, and data
            is the marker data as a (Nt, 3) ndarray.
        """
        data = self._full_marker_data[marker]
        return self.normalize_to_cycle(data, cycle)

    def get_forceplate_data(self, nplate, kind='force', cycle=None):
        """Return forceplate data.

        Parameters
        ----------
        nplate : int
            The forceplate index. Plates are numbered starting from 0, i.e.
            nplate=0 corresponds to Nexus plate 1.
        kind : str
            The type of data to return. Can be 'force', 'moment', or 'cop'
            (center of pressure).

        Returns
        -------
        t_data : tuple
            Tuple of (t, data) where t is the time axis as (Nt,) -shape ndarray,
            and data is the marker data as a (Nt, 3) ndarray.
        """
        if not self._forceplate_data:
            self._forceplate_data = read_data.get_forceplate_data(self.source)
        if nplate < 0 or nplate >= len(self._forceplate_data):
            raise GaitDataError('Invalid plate index %d' % nplate)
        if kind == 'force':
            data = self._forceplate_data[nplate]['F']
        elif kind == 'moment':
            data = self._forceplate_data[nplate]['M']
        elif kind == 'cop':
            data = self._forceplate_data[nplate]['CoP']
        else:
            raise ValueError('Invalid kind of forceplate data requested')
        return self.normalize_analog_to_cycle(data, cycle)

    def get_accelerometer_data(self):
        """Return accelerometer data."""
        raise NotImplementedError

    def _get_fp_events(self):
        """Read the forceplate events."""
        try:
            fp_info = (
                eclipse._eclipse_forceplate_keys(self.eclipse_data)
                if cfg.trial.use_eclipse_fp_info and self.use_eclipse_fp_info
                else None
            )
            # FIXME: marker data already read?
            return utils.detect_forceplate_events(self.source, fp_info=fp_info)
        except GaitDataError:
            logger.warning('Could not detect forceplate events')
            return utils._empty_fp_events()

    def get_cycles(self, cyclespec, max_cycles_per_context=None):
        """Get specified gait cycles from the trial.

        Takes a specification for the desired gait cycles and returns a list of
        Gaitcycle instances.

        Parameters
        ----------
        cyclespec : dict | str | int | tuple | list | None
            The cycles to get. For a context specific cyclespec, it can be dict with
            keys: 'R' and 'L', values: cyclespec as below. If not a dict, the given
            cyclespec will be applied to both contexts.

            For string args: 'all' gets all trial cycles. 'forceplate' gets cycles
            starting with valid forceplate contact. 'unnormalized' (or None) gets a
            Noncycle that is used as a sentinel for unnormalized data.
            If int or a list of int, get the specified cycle(s) from the
            trial. Note that cycle numbering starts from 0.
            A tuple can be used to try different cyclespecs and return the first
            one that has matching cycle. For example, ('forceplate', 0) would return
            forceplate cycles if any, and the first cycle otherwise.

        max_cycles_per_context : int | None
            Maximum number of cycles returned per context.

        Returns
        -------
        list
            List of Gaitcycle instances, sorted by starting frame.
        """

        def _filter_cycles(cycles, context, cyclespec):
            """Takes a list of cycles and filters it according to cyclespec,
            returning only cycles that match the spec"""
            if cyclespec is None:
                return [Noncycle(context=context, trial=self)]
            elif isinstance(cyclespec, int):
                return [cycles[cyclespec]] if cyclespec < len(cycles) else []
            elif isinstance(cyclespec, list):
                return [cycles[c] for c in cyclespec if c < len(cycles)]
            elif cyclespec == 'unnormalized':
                return [Noncycle(context=context, trial=self)]
            elif cyclespec == 'all':
                return cycles
            elif cyclespec == 'forceplate':  # all forceplate cycles
                return [c for c in cycles if c.on_forceplate]
            elif isinstance(cyclespec, tuple):
                # recurse until we have cycles (or cyclespec is exhausted)
                if not cyclespec:
                    return []
                else:
                    return _filter_cycles(
                        cycles, context, cyclespec[0]
                    ) or _filter_cycles(cycles, context, cyclespec[1:])
            else:
                raise TypeError('Invalid argument')

        if not isinstance(cyclespec, dict):
            cyclespec = {'R': cyclespec, 'L': cyclespec}

        cycs_ok = list()
        for context in cyclespec:
            # pick trial cycles for this context
            cycles_ = [c for c in self.cycles if c.context == context.upper()]
            # filter them according to cyclespec
            good_cycles = _filter_cycles(cycles_, context, cyclespec[context])
            if max_cycles_per_context is not None:
                good_cycles = good_cycles[:max_cycles_per_context]
            cycs_ok.extend(good_cycles)

        return sorted(cycs_ok, key=lambda cyc: cyc.start)

    def _scan_cycles(self):
        """Create Gaitcycle instances for this trial.

        Cycle detection is based on trial strike/toeoff markers. To identify
        cycles starting with forceplate contact, the foot strike markers need to
        be matched with forceplate events. A tolerance of STRIKE_TOL is used for
        the matching.
        """
        STRIKE_TOL = 7
        sidestrs = {'R': 'right', 'L': 'left'}
        for strikes in [self.events.lstrikes, self.events.rstrikes]:
            len_s = len(strikes)
            if len_s < 2:
                continue
            if strikes == self.events.lstrikes:
                toeoffs = self.events.ltoeoffs
                context = 'L'
            else:
                toeoffs = self.events.rtoeoffs
                context = 'R'
            for k in range(0, len_s - 1):
                start = strikes[k]
                # see if cycle starts on forceplate strike
                fp_strikes = np.array(self.fp_events[context + '_strikes'])
                if fp_strikes.size == 0:
                    on_forceplate = False
                    plate_idx = None
                else:
                    diffs = np.abs(fp_strikes - start)
                    on_forceplate = min(diffs) <= STRIKE_TOL
                    if on_forceplate:
                        strike_idx = np.argmin(diffs)
                        plate_idx = self.fp_events[context + '_strikes_plate'][
                            strike_idx
                        ]
                    else:
                        plate_idx = None
                    logger.debug(
                        'side %s: cycle start: %d, '
                        'detected fp events: %s' % (context, start, fp_strikes)
                    )
                end = strikes[k + 1]
                toeoff = [x for x in toeoffs if x > start and x < end]
                if len(toeoff) == 0:
                    raise GaitDataError(
                        '%s: no toeoff for cycle starting at '
                        '%d' % (self.trialname, start)
                    )
                elif len(toeoff) > 1:
                    raise GaitDataError(
                        '%s: multiple toeoffs for cycle '
                        'starting at %d' % (self.trialname, start)
                    )
                else:
                    toeoff = toeoff[0]
                fp_str = ' (f)' if on_forceplate else ''
                name = '%s%d%s' % (sidestrs[context], (k + 1), fp_str)
                yield Gaitcycle(
                    start,
                    end,
                    toeoff,
                    context,
                    on_forceplate,
                    plate_idx,
                    self.samplesperframe,
                    trial=self,
                    index=k + 1,
                    name=name,
                )



# -*- coding: utf-8 -*-
"""
Higher level plotting functions (backend agnostic).

@author: Jussi (jnu@iki.fi)
"""

import numpy as np
import os.path as op
import logging

from ..config import cfg
from ..envutils import GaitDataError
from .. import trial, sessionutils, stats, utils
from . import layouts
from .plot_misc import get_backend

logger = logging.getLogger(__name__)


def plot_trials(
    trials,
    layout_name=None,
    backend=None,
    model_normaldata=None,
    emg_normaldata=None,
    cycles=None,
    max_cycles=None,
    emg_mode=None,
    legend_type=None,
    style_by=None,
    color_by=None,
    supplementary_data=None,
    legend=True,
    figtitle=None,
    auto_adjust_emg_layout=False,
):
    """Plot gait trials.

    Create a plot of gait trials using the given layout. Output depends on the
    backend.

    Parameters
    ----------
    trials : list
        List of Trial instances to plot.
    layout_name : str | None
        Name of the plot layout to use (layouts are defined in cfg).
    backend : str | None
        Name of backend to use, currently 'plotly' or 'matplotlib'. If None,
        taken from cfg.
    model_normaldata : dict | None
        Normal data for model variables. If None, taken from cfg.
    emg_normaldata : dict | None
        Normal data for EMG variables. If None, taken from cfg.
    cycles : dict | str | int | tuple | list
        Cycles to plot. See Trial.get_cycles() for details.
    max_cycles : dict | None
        Maximum number of cycles to plot for each variable type. If None, taken
        from cfg.
    emg_mode : str | None
        If 'rms', plot EMG in RMS mode.
    legend_type : str | None
        Legend type for gait cycles (see _get_cycle_name for options). If None,
        taken from cfg.
    style_by : dict | None
        How to style each variable type. If None, taken from cfg.
    color_by : dict | None
        How to color each variable type. If None, taken from cfg.
    supplementary_data : dict | None
        Supplementary data to plot for each variable.
    legend : bool
        Whether to plot legend or not.
    figtitle : str | None
        Main title for the figure.

    Returns
    -------
    fig : Figure | dict
        The figure object. Type depends on backend. Use show_fig() to show it.
    """

    backend_lib = get_backend(backend)
    layout = layouts.get_layout(layout_name)

    if auto_adjust_emg_layout and 'EMG' in layout_name.upper():
        emgs = [tr.emg for tr in trials]
        layout = layouts._rm_dead_channels(emgs, layout)

    return backend_lib.plot_trials(
        trials,
        layout,
        model_normaldata=model_normaldata,
        emg_normaldata=emg_normaldata,
        cycles=cycles,
        max_cycles=max_cycles,
        emg_mode=emg_mode,
        legend_type=legend_type,
        style_by=style_by,
        color_by=color_by,
        supplementary_data=supplementary_data,
        legend=legend,
        figtitle=figtitle,
    )


def _plot_sessions(
    sessions,
    tagged_only=True,
    tags=None,
    layout_name=None,
    backend=None,
    model_normaldata=None,
    cycles=None,
    max_cycles=None,
    emg_mode=None,
    legend_type=None,
    style_by=None,
    color_by=None,
    supplementary_data=None,
    legend=True,
    figtitle=None,
):
    """Gather trials across given sessions and plot them."""
    if not isinstance(sessions, list):
        sessions = [sessions]
    if tags is None:
        tags = cfg.eclipse.tags
    if not tagged_only:
        tags = None
    c3ds_all = sessionutils._get_tagged_dynamic_c3ds_from_sessions(sessions, tags=tags)
    trials = [trial.Trial(c3d) for c3d in c3ds_all]

    return plot_trials(
        trials,
        layout_name=layout_name,
        backend=backend,
        model_normaldata=model_normaldata,
        cycles=cycles,
        max_cycles=max_cycles,
        emg_mode=emg_mode,
        legend_type=legend_type,
        style_by=style_by,
        color_by=color_by,
        supplementary_data=None,
        legend=legend,
        figtitle=figtitle,
    )


def _plot_session_average(
    session,
    layout_name=None,
    tagged_only=True,
    tags=None,
    model_normaldata=None,
    backend=None,
):
    """Average trials from session and plot."""

    layout = layouts.get_layout(layout_name)
    backend_lib = get_backend(backend)

    if tags is None:
        tags = cfg.eclipse.tags
    if not tagged_only:
        tags = None

    c3ds = sessionutils.get_c3ds(session, tags=tags, trial_type='dynamic')
    if not c3ds:
        raise GaitDataError('No dynamic trials found for %s' % session)

    reject_outliers = cfg.trial.outlier_rejection_threshold
    atrial = stats.AvgTrial.from_trials(
        c3ds, reject_outliers=reject_outliers, sessionpath=session
    )
    maintitle_ = '%s (%d trial average)' % (atrial.sessiondir, atrial.nfiles)

    return backend_lib.plot_trials(
        atrial,
        layout,
        model_normaldata=model_normaldata,
        color_by='context',
        figtitle=maintitle_,
    )


def plot_trial_velocities(session, backend=None):
    """Plot median velocities of dynamic trials.

    Parameters
    ----------
    session : str
        Path to session.
    backend : str | None
        Name of backend to use, currently 'plotly' or 'matplotlib'. If None,
        taken from cfg.

    Returns
    -------
    fig : Figure | dict
        The figure object. Type depends on backend. Use show_fig() to show it.
    """
    c3ds = sessionutils.get_c3ds(session, trial_type='dynamic')
    if not c3ds:
        raise GaitDataError('No dynamic trials found for %s' % session)
    labels = [op.splitext(op.split(f)[1])[0] for f in c3ds]
    vels = np.array([utils._trial_median_velocity(tr) for tr in c3ds])
    figtitle = 'Walking speed for %s (average %.2f m/s)' % (
        op.split(session)[-1],
        np.nanmean(vels),
    )
    return get_backend(backend)._plot_vels(vels, labels, title=figtitle)


def plot_trial_timedep_velocities(session, backend=None):
    """Plot time-dependent velocity of dynamic trials in session.

    Parameters
    ----------
    session : str
        Path to session.
    backend : str | None
        Name of backend to use, currently 'plotly' or 'matplotlib'. If None,
        taken from cfg.

    Returns
    -------
    fig : Figure | dict
        The figure object. Type depends on backend. Use show_fig() to show it.

    """
    c3ds = sessionutils.get_c3ds(session, trial_type='dynamic')
    if not c3ds:
        raise GaitDataError('No dynamic trials found for %s' % session)
    vels = list()
    labels = [op.splitext(op.split(f)[1])[0] for f in c3ds]
    # get velocity curves
    for c3d in c3ds:
        _, vel = utils._trial_median_velocity(c3d, return_curve=True)
        # vel = signal.medfilt(vel, 3)  # can be used to filter out spikes
        vels.append(vel)
    figtitle = 'Time-dependent trial velocities for %s' % op.split(session)[-1]
    return get_backend(backend)._plot_timedep_vels(vels, labels, title=figtitle)



# -*- coding: utf-8 -*-
"""

Utility functions for processing gait data.

@author: Jussi (jnu@iki.fi)
"""

from __future__ import division

from builtins import range
from builtins import str
from builtins import zip
from scipy import signal
from matplotlib import path
import matplotlib.pyplot as plt
import numpy as np
import logging

from .envutils import GaitDataError
from .config import cfg
from .numutils import rising_zerocross, digitize_array, falling_zerocross, _baseline


logger = logging.getLogger(__name__)


class TrialEvents(object):
    """A struct-like container for gait event data.
    A dataclass would be better, but requires Python 3.7.
    Logically this might belong in trial.py, but low-level readers (e.g
    nexus.py) need it, so keeping it here makes imports simpler.
    """

    # these are the event types we accept
    event_types = ['rstrikes', 'lstrikes', 'rtoeoffs', 'ltoeoffs', 'general']
    # these are our internal stuff
    secret_stuff = ['_offset']

    def __init__(self, **kwargs):
        self._offset = None
        # init empty lists for all event types
        for evtype in TrialEvents.event_types:
            setattr(self, evtype, list())
        # init event lists via kwargs
        for key, val in kwargs.items():
            setattr(self, key, val)

    def __repr__(self):
        s = '<TrialEvents |'
        for k, evtype in enumerate(TrialEvents.event_types):
            if k > 0:
                s += ','
            ev_vals = getattr(self, evtype)
            s += ' %s: %s' % (evtype, ev_vals)
        s += '>'
        return s

    def __setattr__(self, attr, value):
        if attr in TrialEvents.event_types:  # regular event list
            if not isinstance(value, list):
                raise AttributeError('attribute must be a list')
            else:
                # sort to make sure that event frames are in increasing order
                super(TrialEvents, self).__setattr__(attr, sorted(value))
        elif attr in TrialEvents.secret_stuff:
            super(TrialEvents, self).__setattr__(attr, value)
        else:
            raise AttributeError('%s is not a valid attribute' % attr)

    def subtract_offset(self, offset):
        """Subtract given offset from events."""
        if self._offset:
            raise RuntimeError('offset can be subtracted only once')
        for evtype in TrialEvents.event_types:
            events = getattr(self, evtype)
            events_offset = [e - offset for e in events]
            setattr(self, evtype, events_offset)
        self._offset = offset


def get_contexts(right_first=False):
    """Return the usual contexts and their names as pairs.

    Our default is to return left size first.
    """
    _contexts = [('L', 'Left'), ('R', 'Right')]
    if right_first:
        _contexts.reverse()
    return _contexts


def _empty_fp_events():
    """Container for forceplate events"""
    return dict(
        R_strikes=[],
        R_toeoffs=[],
        L_strikes=[],
        L_toeoffs=[],
        valid=set(),
        R_strikes_plate=[],
        L_strikes_plate=[],
        our_fp_info={},
        coded='',
    )


def marker_gaps(mdata, ignore_edge_gaps=True):
    """Find gaps for a marker.

    Parameters
    ----------
    mdata : ndarray
        Marker position data. See read_data.get_marker_data()
    ignore_edge_gaps : bool, optional
        If True, leading and trailing gaps are ignored.

    Returns
    -------
    ndarray
        The indices of frames where gaps occur.
    """
    allzero = np.any(mdata, axis=1).astype(int)
    if ignore_edge_gaps:
        nleading = allzero.argmax()
        allzero_trim = np.trim_zeros(allzero)
        gap_inds = np.where(allzero_trim == 0)[0] + nleading
    else:
        gap_inds = np.where(allzero == 0)[0]
    return gap_inds


def _step_width(source):
    """Compute step width over trial cycles.

    For details of computation, see:
    https://www.vicon.com/faqs/software/how-does-nexus-plug-in-gait-and-polygon-calculate-gait-cycle-parameters-spatial-and-temporal
    Returns context keyed dict of lists.
    FIXME: marker name into params?
    FIXME: this (and similar) may also need to take Trial instance as argument
    to avoid creating new Trials
    """
    from .trial import Trial

    tr = Trial(source)
    sw = dict()
    mkr = 'TOE'  # marker name without context
    mkrdata = tr._full_marker_data
    # FIXME: why not use cycles here?
    for context, strikes in zip(['L', 'R'], [tr.events.lstrikes, tr.events.rstrikes]):
        sw[context] = list()
        nstrikes = len(strikes)
        if nstrikes < 2:
            continue
        # contralateral vars
        context_co = 'L' if context == 'R' else 'R'
        strikes_co = tr.events.lstrikes if context == 'R' else tr.events.rstrikes
        mname = context + mkr
        mname_co = context_co + mkr
        for j, strike in enumerate(strikes):
            if strike == strikes[-1]:  # last strike on this side
                break
            pos_this = mkrdata[mname][strike]
            pos_next = mkrdata[mname][strikes[j + 1]]
            strikes_next_co = [k for k in strikes_co if k > strike]
            if len(strikes_next_co) == 0:  # no subsequent contralateral strike
                break
            pos_next_co = mkrdata[mname_co][strikes_next_co[0]]
            # vector distance between 'step lines' (see url above)
            V1 = pos_next - pos_this
            V1 /= np.linalg.norm(V1)
            VC = pos_next_co - pos_this
            VCP = V1 * np.dot(VC, V1)  # proj to ipsilateral line
            VSW = VCP - VC
            # marker data is in mm, but return step width in m
            sw[context].append(np.linalg.norm(VSW) / 1000.0)
    return sw


def avg_markerdata(mkrdata, markers, roi=None, fail_on_gaps=True, avg_velocity=False):
    """Average marker data.

    Parameters
    ----------
    mkrdata : dict
        The markerdata dict.
    markers : list
        Markers to average.
    roi : array-like
        If given, specified a ROI. Gaps outside the ROI will be ignored.
    fail_on_gaps : bool, optional
        If True, raise an exception on ANY gaps. Otherwise, markers with gaps
        will not be included in the average.
    avg_velocity : bool, optional
        If True, return averaged marker velocity data instead of position.

    Returns
    -------
    ndarray
        The averaged data (Nx3).
    """
    data_shape = mkrdata[markers[0]].shape
    mdata_avg = np.zeros(data_shape)
    if roi is None:
        roi = [0, data_shape[0]]
    roi_frames = np.arange(roi[0], roi[1])
    n_ok = 0
    for marker in markers:
        mdata = mkrdata[marker]
        if avg_velocity:
            mdata = np.gradient(mdata)[0]
        gap_frames = marker_gaps(mdata)
        if np.intersect1d(roi_frames, gap_frames).size > 0:
            if fail_on_gaps:
                raise GaitDataError('Averaging data for %s has gaps' % marker)
            else:
                logger.warning(
                    'marker %s cannot be included in average due to gaps' % marker
                )
                continue
        else:
            mdata_avg += mdata
            n_ok += 1
    if n_ok == 0:
        raise GaitDataError('all markers have gaps, cannot average')
    else:
        return mdata_avg / n_ok


# FIXME: marker sets could be moved into models.py?
def _pig_markerset(fullbody=True, sacr=True):
    """ PiG marker set as dict (empty values) """
    _pig = [
        'LASI',
        'RASI',
        'LTHI',
        'LKNE',
        'LTIB',
        'LANK',
        'LHEE',
        'LTOE',
        'RTHI',
        'RKNE',
        'RTIB',
        'RANK',
        'RHEE',
        'RTOE',
    ]
    if fullbody:
        _pig += [
            'LFHD',
            'RFHD',
            'LBHD',
            'RBHD',
            'C7',
            'T10',
            'CLAV',
            'STRN',
            'RBAK',
            'LSHO',
            'LELB',
            'LWRA',
            'LWRB',
            'LFIN',
            'RSHO',
            'RELB',
            'RWRA',
            'RWRB',
            'RFIN',
        ]
    # add pelvis posterior markers; SACR or RPSI/LPSI
    _pig.extend(['SACR'] if sacr else ['RPSI', 'LPSI'])
    return {mkr: None for mkr in _pig}


def _pig_pelvis_markers():
    return ['RASI', 'RPSI', 'LASI', 'LPSI', 'SACR']


def is_plugingait_set(mkrdata):
    """Check whether marker data set corresponds to Plug-in Gait (full body or
    lower body only). Extra markers are accepted."""
    mkrs = set(mkrdata.keys())
    # required markers
    lb_mkrs_sacr = set(_pig_markerset(fullbody=False).keys())
    lb_mkrs_psi = set(_pig_markerset(fullbody=False, sacr=False).keys())
    return lb_mkrs_psi.issubset(mkrs) or lb_mkrs_sacr.issubset(mkrs)


def _check_markers_flipped(mkrdata):
    """Check for markers that may typically get flipped in labeling.
    Yields pairs of flipped markers"""
    MAX_ANGLE = 90  # max angle to consider vectors 'similarly oriented'
    # HEE-TOE checks
    for side in ['L', 'R']:
        # compare HEE-TOE line to pelvis orientation
        mkr_toe, mkr_hee = '%sTOE' % side, '%sHEE' % side
        ht = _normalize(mkrdata[mkr_toe] - mkrdata[mkr_hee])
        if side + 'PSI' in mkrdata:
            pa = _normalize(mkrdata[side + 'ASI'] - mkrdata[side + 'PSI'])
        elif 'SACR' in mkrdata:
            pa = _normalize(mkrdata[side + 'ASI'] - mkrdata['SACR'])
        angs = np.arccos(np.sum(ht * pa, axis=1)) / np.pi * 180
        if np.nanmedian(angs) > MAX_ANGLE:
            yield mkr_toe, mkr_hee


def _principal_movement_direction(mdata):
    """Return principal movement direction (dimension of maximum variance)"""
    inds_ok = np.where(np.any(mdata, axis=1))  # make sure that gaps are ignored
    return np.argmax(np.var(mdata[inds_ok], axis=0))


def _get_foot_contact_vel(mkrdata, fp_events, medians=True, roi=None):
    """Return foot velocities during forceplate strike/toeoff frames.
    fp_events is from detect_forceplate_events() If medians=True, return median
    values."""
    results = dict()
    for context, markers in zip(
        ('R', 'L'), [cfg.autoproc.right_foot_markers, cfg.autoproc.left_foot_markers]
    ):
        # we can be less strict about gaps here, since missing a marker or two
        # probably won't really affect the velocity
        footctrv_ = avg_markerdata(
            mkrdata,
            markers,
            roi=roi,
            fail_on_gaps=False,
            avg_velocity=True,
        )
        footctrv = np.linalg.norm(footctrv_, axis=1)
        strikes = fp_events[context + '_strikes']
        toeoffs = fp_events[context + '_toeoffs']
        results[context + '_strike'] = footctrv[strikes]
        results[context + '_toeoff'] = footctrv[toeoffs]
    if medians:
        results = {
            key: (np.array([np.median(x)]) if x.size > 0 else x)
            for key, x in results.items()
        }
    return results


def _get_foot_swing_velocity(footctrv, max_peak_velocity, min_swing_velocity):
    """Compute foot swing velocity from scalar velocity data (footctrv)"""
    # find maxima of velocity: derivative crosses zero and values ok
    vd = np.gradient(footctrv)
    vdz_ind = falling_zerocross(vd)
    inds = np.where(footctrv[vdz_ind] < max_peak_velocity)[0]
    if len(inds) == 0:
        raise GaitDataError('Cannot find acceptable velocity peaks')
    # delete spurious peaks (where min swing velocity is not attained)
    vs = footctrv[vdz_ind[inds]]
    not_ok = np.where(vs < vs.max() * min_swing_velocity)
    vs = np.delete(vs, not_ok)
    return np.median(vs)


def _normalize(V):
    """Normalize rows of matrix V"""
    Vn = np.linalg.norm(V, axis=1)
    # quietly return all nans for length zero vectors
    with np.errstate(divide='ignore', invalid='ignore'):
        return V / Vn[:, np.newaxis]


def _get_foot_points(mkrdata, context, footlen=None):
    """Estimate points in the xy plane enclosing the foot.

    Foot is modeled as a triangle with three points: heel, lateral leading edge
    (small toe) and medial leading edge (hallux)"""
    # marker data as N x 3 matrices
    heeP = mkrdata[context + 'HEE']
    toeP = mkrdata[context + 'TOE']
    ankP = mkrdata[context + 'ANK']
    # heel - toe vectors
    htV = toeP - heeP
    htVn = _normalize(htV)
    # heel - ankle vectors
    haV = ankP - heeP
    ha_len = np.linalg.norm(haV, axis=1)
    # end point of foot (just beyond 2nd toe)
    if footlen is None:
        # rough estimate based on marker distances
        logger.debug('using estimated foot length')
        footlen = np.median(ha_len) * cfg.autoproc.foot_relative_len
    foot_end = heeP + htVn * footlen
    # projection of HEE-ANK to HEE-TOE line
    ha_htV = htVn * np.sum(haV * htVn, axis=1)[:, np.newaxis]
    # lateral ANK vector (from HEE-TOE line to ANK)
    lankV = haV - ha_htV
    # edge points are coplanar with markers but not with the foot
    # lateral foot edge
    lat_edge = foot_end + lankV
    # medial foot edge
    med_edge = foot_end - lankV
    # heel edge (compensate for marker diameter)
    heel_edge = heeP + htVn * cfg.autoproc.marker_diam / 2
    logger.debug(
        'foot length: %.1f mm width: %.1f mm'
        % (
            np.nanmedian(np.linalg.norm(heel_edge - foot_end, axis=1)),
            np.nanmedian(np.linalg.norm(lat_edge - med_edge, axis=1)),
        )
    )
    return {'heel': heel_edge, 'lateral': lat_edge, 'medial': med_edge, 'toe': foot_end}


def _leading_foot(mkrdata, roi=None):
    """Determine which foot is leading (ahead in the direction of gait).
    Returns n-length list of 'R' or 'L' correspondingly (n = number of
    frames). Gaps are indicated as None. mkrdata must include foot and
    pelvis markers"""
    subj_pos = avg_markerdata(
        mkrdata, cfg.autoproc.track_markers, roi=roi, fail_on_gaps=False
    )
    # FIXME: should not use a single dim here
    gait_dim = _principal_movement_direction(subj_pos)
    gait_dir = np.median(np.diff(subj_pos, axis=0), axis=0)[gait_dim]
    lfoot = avg_markerdata(
        mkrdata, cfg.autoproc.left_foot_markers, roi=roi, fail_on_gaps=False
    )[:, gait_dim]
    rfoot = avg_markerdata(
        mkrdata, cfg.autoproc.right_foot_markers, roi=roi, fail_on_gaps=False
    )[:, gait_dim]
    cmpfun = np.greater if gait_dir > 0 else np.less
    return [
        None if R == 0.0 or L == 0.0 else ('R' if cmpfun(R, L) else 'L')
        for R, L in zip(rfoot, lfoot)
    ]


def _trial_median_velocity(source, return_curve=False):
    """Compute median velocity (walking speed) over whole trial by
    differentiation of marker data from track markers. Up/down movement of
    markers may slightly increase speed compared to time-distance values.
    If return_curve, return velocity curve normalized to 0..100% of trial"""
    from . import read_data

    try:
        frate = read_data.get_metadata(source)['framerate']
        mkrdata = read_data.get_marker_data(source, cfg.autoproc.track_markers)
        vel_3 = avg_markerdata(mkrdata, cfg.autoproc.track_markers, avg_velocity=True)
        vel_ = np.sqrt(np.sum(vel_3 ** 2, 1))  # scalar velocity
    except (GaitDataError, ValueError):
        if return_curve:
            nanvec = np.empty((100, 1))
            nanvec[:] = np.nan
            return np.nan, nanvec
        else:
            return np.nan
    vel = np.median(vel_[np.where(vel_)])  # ignore zeros
    vel_ms = vel * frate / 1000.0  # convert to m/s
    if return_curve:
        tn = np.linspace(0, 100, 101)
        vel_curve = np.interp(tn, np.linspace(0, 100, len(vel_)), vel_)
        return vel_ms, vel_curve * frate / 1000.0
    else:
        return vel_ms


def _point_in_poly(poly, pt):
    """Point-in-polygon. poly is ordered nx3 array of vertices and P is mx3
    array of points. Returns mx3 array of booleans. 3rd dim is currently
    ignored"""
    p = path.Path(poly[:, :2])
    return p.contains_point(pt)


def detect_forceplate_events(source, mkrdata=None, fp_info=None, roi=None):
    """Detect frames where valid forceplate strikes and toeoffs occur.
    Uses forceplate data and estimated foot shape.

    If supplied, mkrdata must include foot and pelvis markers. Otherwise
    it will be read.

    If fp_info dict is supplied, no marker-based checks will be done;
    instead the Eclipse forceplate info will be used to determine the foot.
    Eclipse info is written e.g. as {FP1: 'Left'} where plate indices start
    from 1 and the value can be 'Auto', 'Left', 'Right' or 'Invalid'.
    Even if Eclipse info is used, foot strike and toeoff frames must be
    determined from forceplate data.

    If roi is given e.g. [100, 300], all marker data checks will be restricted
    to roi.
    """

    def _foot_plate_check(fpdata, mkrdata, fr0, side, footlen):
        """Helper for foot-plate check.

        Returns 0, 1, 2 for: completely outside plate, partially outside plate,
        inside plate, respectively.
        """
        allpts = _get_foot_points(mkrdata, side, footlen)
        poly = fpdata['plate_corners']
        pts_ok = list()
        for label, pts in allpts.items():
            pt = pts[fr0, :]
            pt_ok = _point_in_poly(poly, pt)
            logger.debug('%s point %son plate' % (label, '' if pt_ok else 'not '))
            pts_ok.append(pt_ok)
        if all(pts_ok):
            return 2
        elif any(pts_ok):
            return 1
        else:
            return 0

    def _threshold_forceplate(fp, bodymass=None):
        """Get candidate foot strike and toeoff frames by considering force only"""
        # apply median filter to remove spikes
        # XXX: kernel size should maybe depend on sampling freq?
        forcetot = signal.medfilt(fp['Ftot'], kernel_size=3)
        forcetot = _baseline(forcetot)
        fmaxind = np.argmax(forcetot)
        fmax = forcetot[fmaxind]
        logger.debug('peak force: %.2f N at sample %d' % (fmax, fmaxind))

        # find force threshold
        if bodymass is None:
            f_threshold = cfg.autoproc.forceplate_contact_threshold * fmax
            logger.warning(
                'body mass unknown, thresholding force at %.2f N', f_threshold
            )
        else:
            logger.debug('body mass %.2f kg' % bodymass)
            f_threshold = cfg.autoproc.forceplate_contact_threshold * bodymass * 9.81
            if fmax < cfg.autoproc.forceplate_min_weight * bodymass * 9.81:
                logger.debug('insufficient max. force on plate')
                force_checks_ok = False

        # find the indices around peak where force crosses threshold
        f_rising_inds = rising_zerocross(forcetot - f_threshold)
        f_falling_inds = falling_zerocross(forcetot - f_threshold)
        try:
            f_rising_ind = f_rising_inds[np.where(f_rising_inds < fmaxind)][-1]
            f_falling_ind = f_falling_inds[np.where(f_falling_inds > fmaxind)][0]
            logger.debug('force rise: %d fall: %d' % (f_rising_ind, f_falling_ind))
            # these are 0-based frame indices (=1 less than Nexus frame index)
            strike_fr = int(np.round(f_rising_ind / info['samplesperframe']))
            toeoff_fr = int(np.round(f_falling_ind / info['samplesperframe']))
            logger.debug(
                'foot strike at frame %d, toeoff at %d' % (strike_fr, toeoff_fr)
            )
            force_checks_ok = True
        except IndexError:
            logger.debug('cannot detect force rise/fall')
            force_checks_ok = False
            strike_fr, toeoff_fr = None, None
        return strike_fr, toeoff_fr, force_checks_ok

    def _context_from_eclipse(fp_info, plate):
        """Interpret context from Eclipse data.
        Returns tuple of (context, detect_context), where context is 'R', 'L'
        (or None for invalid/unknown) and detect_context indicates whether we
        should do our own autodetection.
        """
        context = None
        if fp_info is not None and plate in fp_info:
            eclipse_context = fp_info[plate]
            logger.debug('using Eclipse forceplate info: %s' % eclipse_context)
            if eclipse_context == 'Right':
                detect_context = False
                context = 'R'
            elif eclipse_context == 'Left':
                detect_context = False
                context = 'L'
            elif eclipse_context == 'Invalid':
                detect_context = False
            elif eclipse_context == 'Auto':
                detect_context = True
            else:
                raise GaitDataError('unexpected Eclipse forceplate field')
        else:
            logger.debug('not using Eclipse forceplate info')
            detect_context = True
        return context, detect_context

    from . import read_data

    logger.debug('detecting forceplate events from %s' % source)
    results = _empty_fp_events()

    # get subject info
    info = read_data.get_metadata(source)
    fpdata = read_data.get_forceplate_data(source)
    if cfg.autoproc.nexus_forceplate_devnames:
        logger.info(
            'using configured plates: %s' % cfg.autoproc.nexus_forceplate_devnames
        )
    if not fpdata:
        logger.warning('no forceplates')
        return results
    footlen = info['subj_params']['FootLen']
    rfootlen = info['subj_params']['RFootLen']
    lfootlen = info['subj_params']['LFootLen']
    if footlen is not None:
        logger.debug('(obsolete) single foot length parameter set to %.2f' % footlen)
        rfootlen = lfootlen = footlen
    elif rfootlen is not None and lfootlen is not None:
        logger.debug(
            'foot length parameters set to r=%.2f, l=%.2f' % (rfootlen, lfootlen)
        )
    else:
        logger.debug('foot length parameter not set')
    bodymass = info['subj_params']['Bodymass']
    if mkrdata is None:  # not supplied as parameter
        mkrs = (
            cfg.autoproc.right_foot_markers
            + cfg.autoproc.left_foot_markers
            + cfg.autoproc.track_markers
        )
        mkrdata = read_data.get_marker_data(source, mkrs)

    logger.debug('acquiring gait events')
    events_0 = automark_events(source, mkrdata=mkrdata, mark=False, roi=roi)

    # loop over the plates; our internal forceplate index is 0-based
    for plate_ind, fp in enumerate(fpdata):
        logger.debug('analyzing plate %d' % plate_ind)
        # XXX: are we sure that the plate indices always match Eclipse?
        plate = 'FP' + str(plate_ind + 1)  # Eclipse numbering starts from FP1
        context, detect_context = _context_from_eclipse(fp_info, plate)

        strike_fr, toeoff_fr, force_checks_ok = _threshold_forceplate(fp, bodymass)
        if not force_checks_ok:
            context = None

        # check foot markers (or points) to determine context and validity
        if force_checks_ok and detect_context:
            logger.debug('autodetecting context')
            # allows foot to settle for 50 ms after strike
            settle_fr = int(50 / 1000 * info['framerate'])
            fr0 = strike_fr + settle_fr
            # context is determined by leading foot at strike time
            this_context = _leading_foot(mkrdata, roi=roi)[fr0]
            if this_context is None:
                raise GaitDataError('cannot determine leading foot from marker data')
            footlen = rfootlen if this_context == 'R' else lfootlen
            logger.debug('checking contact for leading foot: %s' % this_context)
            foot_contacts_ok = (
                _foot_plate_check(fp, mkrdata, fr0, this_context, footlen) == 2
            )
            # to eliminate double contacts, check that contralateral foot is not on plate
            # this needs marker-based events
            if foot_contacts_ok and events_0 is not None:
                contra_side = 'R' if this_context == 'L' else 'L'
                contra_strikes = events_0[contra_side + '_strikes']
                contra_strikes_next = contra_strikes[
                    np.where(contra_strikes > strike_fr)
                ]
                if contra_strikes_next.size == 0:
                    logger.debug('no following contralateral strike')
                else:
                    fr0 = contra_strikes_next[0] + settle_fr
                    logger.debug(
                        'checking the following contralateral strike '
                        '(at frame %d)' % fr0
                    )
                    contra_next_ok = (
                        _foot_plate_check(fp, mkrdata, fr0, contra_side, footlen) == 0
                    )
                    foot_contacts_ok &= contra_next_ok
                contra_strikes_prev = contra_strikes[
                    np.where(contra_strikes < strike_fr)
                ]
                if contra_strikes_prev.size == 0:
                    logger.debug('no previous contralateral strike')
                else:
                    fr0 = contra_strikes_prev[-1] + settle_fr
                    logger.debug(
                        'checking previous contact for contralateral '
                        'foot (at frame %d)' % fr0
                    )
                    contra_prev_ok = (
                        _foot_plate_check(fp, mkrdata, fr0, contra_side, footlen) == 0
                    )
                    foot_contacts_ok &= contra_prev_ok
            context = this_context if foot_contacts_ok else None

        if context:
            logger.debug(
                'plate %d: valid foot strike on %s at frame %d'
                % (plate_ind, context, strike_fr)
            )
            results['valid'].add(context)
            results[context + '_strikes'].append(strike_fr)
            results[context + '_toeoffs'].append(toeoff_fr)
            results[context + '_strikes_plate'].append(plate_ind)
            results['our_fp_info'][plate] = 'Right' if context == 'R' else 'Left'
            results['coded'] += context
        else:
            logger.debug('plate %d: no valid foot strike' % plate_ind)
            results['our_fp_info'][plate] = 'Invalid'
            results['coded'] += 'X'

    logger.debug(results)
    return results


def automark_events(
    source,
    mkrdata=None,
    events_range=None,
    fp_events=None,
    vel_thresholds=None,
    roi=None,
    start_on_forceplate=False,
    plot=False,
    mark=True,
):
    """Automatically mark foot strike and toeoff events.

    Events are marked based on velocity thresholding. Absolute thresholds can be
    specified as arguments. Otherwise, relative thresholds will be calculated
    based on the data. Optimal results will be obtained when thresholds are
    predetermined based on forceplate data, but it is not necessary.

    Before running automark, run reconstruct, label, gap fill and filter
    pipelines. Filtering is important to get reasonably smooth derivatives.

    Parameters
    ----------
    source : str | ViconNexus
        The data source, either c3d filename or ViconNexus connection. For Nexus
        connections, the events can automatically be inserted into Nexus. For c3d
        files, the events are returned but not actually written to the c3d file.
    mkrdata : dict, optional
        The marker data dict. If not given, it will be read from the source. If
        given, it must include foot markers and subject tracking markers (see cfg).
    events_range : array_like, optional
        If specified, the events will be restricted to given coordinate range in
        the principal gait direction. E.g. [-1000, 1000]
    fp_events : dict, optional
        If not None, specifies the forceplate detected strikes and toeoffs
        (see utils.detect_forceplate_events). These will be then considered the
        ground truth and will replace nearby autodetected events.
    vel_thresholds : dict, optional
        Absolute velocity thresholds for identifying events. These can be obtained
        from forceplate data (utils.check_forceplate_contact). If None, relative
        thresholds will be computed based on marker data.
    roi : array_like, optional
        If not None, specifies a ROI (in frames) inside which to mark events.
    start_on_forceplate : bool, optional
        If True, try to start the first gait cycle on forceplate, i.e. events
        earlier than the first forceplate contact will not be marked.
    plot : bool, optional
        Plot velocity curves and events using matplotlib. Mostly for debug purposes.
    mark : bool, optional
        If False, do not actually insert the marker events into Nexus.
    """

    from .read_data import get_metadata, get_marker_data

    info = get_metadata(source)
    frate = info['framerate']
    # some operations are Nexus specific and make no sense for c3d source
    from . import nexus

    # TODO: move into config
    # marker data is assumed to be in mm
    # mm/frame = 1000 m/frame = 1000/frate m/s
    VEL_CONV = 1000 / frate
    # reasonable limit for peak velocity (m/s before multiplier)
    MAX_PEAK_VELOCITY = 12 * VEL_CONV
    # reasonable limits for velocity on slope (increasing/decreasing)
    MAX_SLOPE_VELOCITY = 6 * VEL_CONV
    MIN_SLOPE_VELOCITY = 0  # not currently in use
    # minimum swing velocity (rel to max velocity)
    MIN_SWING_VELOCITY = 0.5
    # median prefilter width
    PREFILTER_MEDIAN_WIDTH = 3
    # tolerance for matching forceplate and vel. thresholded events
    FP_EVENT_TOL = 10

    if vel_thresholds is None:
        vel_thresholds = {
            'L_strike': None,
            'L_toeoff': None,
            'R_strike': None,
            'R_toeoff': None,
        }

    if mkrdata is None:
        # FIXME: missing markers are not detected here?
        reqd_markers = (
            cfg.autoproc.right_foot_markers
            + cfg.autoproc.left_foot_markers
            + cfg.autoproc.track_markers
        )
        mkrdata = get_marker_data(source, reqd_markers)

    rfootctrv_ = avg_markerdata(
        mkrdata,
        cfg.autoproc.right_foot_markers,
        roi=roi,
        fail_on_gaps=False,
        avg_velocity=True,
    )
    rfootctrv = np.linalg.norm(rfootctrv_, axis=1)
    lfootctrv_ = avg_markerdata(
        mkrdata,
        cfg.autoproc.left_foot_markers,
        roi=roi,
        fail_on_gaps=False,
        avg_velocity=True,
    )
    lfootctrv = np.linalg.norm(lfootctrv_, axis=1)

    # position data: use ANK marker
    rfootctrP = mkrdata['RANK']
    lfootctrP = mkrdata['LANK']

    strikes_all = {}
    toeoffs_all = {}

    # loop: same operations for left / right foot
    for context, footctrP, footctrv in zip(
        ('R', 'L'), (rfootctrP, lfootctrP), (rfootctrv, lfootctrv)
    ):
        logger.debug('marking side %s' % context)
        # foot center position
        # filter scalar velocity data to suppress noise and spikes
        footctrv = signal.medfilt(footctrv, PREFILTER_MEDIAN_WIDTH)
        # get peak (swing) velocity
        maxv = _get_foot_swing_velocity(footctrv, MAX_PEAK_VELOCITY, MIN_SWING_VELOCITY)

        # compute thresholds
        if cfg.autoproc.use_fp_vel_thresholds and vel_thresholds[context + '_strike']:
            threshold_fall_ = vel_thresholds[context + '_strike']
        else:
            threshold_fall_ = maxv * cfg.autoproc.strike_vel_threshold
        if cfg.autoproc.use_fp_vel_thresholds and vel_thresholds[context + '_toeoff']:
            threshold_rise_ = vel_thresholds[context + '_toeoff']
        else:
            threshold_rise_ = maxv * cfg.autoproc.toeoff_vel_threshold
        logger.debug(
            'side: %s, default thresholds fall/rise: %.2f/%.2f'
            % (
                context,
                maxv * cfg.autoproc.strike_vel_threshold,
                maxv * cfg.autoproc.toeoff_vel_threshold,
            )
        )
        logger.debug('using thresholds: %.2f/%.2f' % (threshold_fall_, threshold_rise_))
        # find point where velocity crosses threshold
        # foot strikes (velocity decreases)
        cross = falling_zerocross(footctrv - threshold_fall_)
        # exclude edges of data vector
        fmax = len(footctrv) - 1
        cross = cross[np.where(np.logical_and(cross > 0, cross < fmax))]
        # check velocity on slope
        cind_min = np.logical_and(
            footctrv[cross - 1] < MAX_SLOPE_VELOCITY,
            footctrv[cross - 1] > MIN_SLOPE_VELOCITY,
        )
        cind_max = np.logical_and(
            footctrv[cross + 1] < MAX_SLOPE_VELOCITY,
            footctrv[cross + 1] > MIN_SLOPE_VELOCITY,
        )
        strikes = cross[np.logical_and(cind_min, cind_max)]

        # check for foot swing (velocity maximum) between consecutive strikes
        # if no swing, keep deleting the latter event until swing is found
        bad = []
        for sind in range(len(strikes)):
            if sind in bad:
                continue
            for sind2 in range(sind + 1, len(strikes)):
                swing_max_vel = footctrv[strikes[sind] : strikes[sind2]].max()
                # logger.debug('check %d-%d' % (strikes[sind], strikes[sind2]))
                if swing_max_vel < maxv * MIN_SWING_VELOCITY:
                    logger.debug(
                        'no swing between strikes %d-%d, deleting %d'
                        % (strikes[sind], strikes[sind2], strikes[sind2])
                    )
                    bad.append(sind2)
                else:
                    break
        strikes = np.delete(strikes, bad)

        if len(strikes) == 0:
            raise GaitDataError('No valid foot strikes detected')

        # toe offs (velocity increases)
        cross = rising_zerocross(footctrv - threshold_rise_)
        cross = cross[np.where(np.logical_and(cross > 0, cross < len(footctrv)))]
        cind_min = np.logical_and(
            footctrv[cross - 1] < MAX_SLOPE_VELOCITY,
            footctrv[cross - 1] > MIN_SLOPE_VELOCITY,
        )
        cind_max = np.logical_and(
            footctrv[cross + 1] < MAX_SLOPE_VELOCITY,
            footctrv[cross + 1] > MIN_SLOPE_VELOCITY,
        )
        toeoffs = cross[np.logical_and(cind_min, cind_max)]

        if len(toeoffs) == 0:
            raise GaitDataError('Could not detect any toe-off events')

        # check for multiple toeoffs
        for s1, s2 in list(zip(strikes, np.roll(strikes, -1)))[:-1]:
            to_this = np.where(np.logical_and(toeoffs > s1, toeoffs < s2))[0]
            if len(to_this) > 1:
                logger.debug(
                    '%d toeoffs during cycle, keeping the last one' % len(to_this)
                )
                toeoffs = np.delete(toeoffs, to_this[:-1])

        logger.debug('autodetected strike events: %s' % strikes)
        logger.debug('autodetected toeoff events: %s' % toeoffs)

        # select events for which the foot is close enough to center frame
        if events_range:
            mdata = avg_markerdata(mkrdata, cfg.autoproc.track_markers, roi=roi)
            fwd_dim = _principal_movement_direction(mdata)
            strike_pos = footctrP[strikes, fwd_dim]
            dist_ok = np.logical_and(
                strike_pos > events_range[0], strike_pos < events_range[1]
            )
            # exactly zero position at strike should indicate a gap -> exclude
            # TODO: smarter gap handling
            dist_ok = np.logical_and(dist_ok, strike_pos != 0)
            strikes = strikes[dist_ok]

        # correct foot strikes with force plate autodetected events
        if fp_events and fp_events[context + '_strikes']:
            fp_strikes = fp_events[context + '_strikes']
            logger.debug('forceplate strikes: %s' % fp_strikes)
            # find best fp matches for all strikes
            fpc = digitize_array(strikes, fp_strikes)
            ok_ind = np.where(np.abs(fpc - strikes) < FP_EVENT_TOL)[0]
            if ok_ind.size == 0:
                logger.warning(
                    'could not match forceplate strike with an autodetected strike'
                )
            else:
                # replace with fp detected strikes
                strikes[ok_ind] = fpc[ok_ind]
                logger.debug('fp corrected strikes: %s' % strikes)
            # toeoffs
            fp_toeoffs = fp_events[context + '_toeoffs']
            logger.debug('forceplate toeoffs: %s' % fp_toeoffs)
            fpc = digitize_array(toeoffs, fp_toeoffs)
            ok_ind = np.where(np.abs(fpc - toeoffs) < FP_EVENT_TOL)[0]
            if ok_ind.size == 0:
                logger.warning(
                    'could not match forceplate toeoff with an autodetected toeoff'
                )
            else:
                toeoffs[ok_ind] = fpc[ok_ind]
                logger.debug('fp corrected toeoffs: %s' % toeoffs)
            # delete strikes before (actual) 1st forceplate contact
            if start_on_forceplate and len(fp_strikes) > 0:
                # use a tolerance here to avoid deleting possible (uncorrected)
                # strike near the fp
                not_ok = np.where(strikes < fp_strikes[0] - FP_EVENT_TOL)[0]
                if not_ok.size > 0:
                    logger.debug(
                        'deleting foot strikes before forceplate: %s' % strikes[not_ok]
                    )
                    strikes = np.delete(strikes, not_ok)

        if roi is not None:
            strikes = np.extract(
                np.logical_and(roi[0] <= strikes + 1, strikes + 1 <= roi[1]), strikes
            )

            toeoffs = np.extract(
                np.logical_and(roi[0] <= toeoffs + 1, toeoffs + 1 <= roi[1]), toeoffs
            )

        if len(strikes) == 0:
            raise GaitDataError('No valid foot strikes detected')

        # delete toeoffs that are not between strike events
        not_ok = np.where(
            np.logical_or(toeoffs <= min(strikes), toeoffs >= max(strikes))
        )
        toeoffs = np.delete(toeoffs, not_ok)

        logger.debug('final strike events: %s' % strikes)
        logger.debug('final toeoff events: %s' % toeoffs)

        if mark:
            if not nexus._is_vicon_instance(source):
                raise ValueError('event marking supported only for Nexus')
            vicon = nexus.viconnexus()
            nexus._create_events(vicon, context, strikes, toeoffs)

        strikes_all[context] = strikes
        toeoffs_all[context] = toeoffs

        # plot velocities w/ thresholds and marked events
        if plot:
            first_call = context == 'R'
            if first_call:
                f, (ax1, ax2) = plt.subplots(2, 1)
            ax = ax1 if first_call else ax2
            ax.plot(footctrv, 'g', label='foot center velocity ' + context)
            # algorithm, fixed thresholds
            ax.plot(strikes, footctrv[strikes], 'kD', markersize=10, label='strike')
            ax.plot(toeoffs, footctrv[toeoffs], 'k^', markersize=10, label='toeoff')
            ax.legend(numpoints=1, fontsize=10)
            ax.set_ylim(0, maxv + 10)
            if not first_call:
                plt.xlabel('Frame')
            ax.set_ylabel('Velocity (mm/frame)')
            ax.set_title('Left' if context == 'L' else 'Right')

    if plot:
        plt.show()

    return {
        'R_strikes': strikes_all['R'],
        'L_strikes': strikes_all['L'],
        'R_toeoffs': toeoffs_all['R'],
        'L_toeoffs': toeoffs_all['L'],
    }



# -*- coding: utf-8 -*-
"""
matplotlib based plotting functions

@author: Jussi (jnu@iki.fi)
"""

from __future__ import division

from builtins import zip
from builtins import range
from functools import partial
from collections import OrderedDict, defaultdict
import itertools
import matplotlib
from matplotlib.figure import Figure
import matplotlib.gridspec as gridspec
import plotly.graph_objs as go
import plotly.io as pio
import numpy as np
import logging
import io

from .plot_common import (
    _get_cycle_name,
    _var_title,
    _handle_cyclespec,
    _cyclical_mapper,
    _handle_style_and_color_args,
    _color_by_params,
    _style_by_params,
    _emg_yscale,
    _get_trial_cycles,
    _triage_var,
    _compose_varname,
    _nested_get,
    _var_unit,
)
from .. import models, normaldata, utils
from ..config import cfg
from ..envutils import GaitDataError
from ..stats import AvgTrial
from ..timedist import _pick_common_vars
from . import layouts

logger = logging.getLogger(__name__)


def _plot_extracted_table(curve_vals, vardefs):
    """Plot comparison of extracted gait curve values as a table."""
    contexts = utils.get_contexts()
    col_labels = list(curve_vals.keys())
    row_labels = [_compose_varname(vardef) for vardef in vardefs]
    table = list()
    for vardef in vardefs:
        row = list()
        for session_vals in curve_vals.values():
            element = ''
            for ctxt, _ in contexts:
                vardef_ctxt = [ctxt + vardef[0]] + vardef[1:]
                if vardef_ctxt[0] not in session_vals:
                    logger.debug(
                        '%s was not collected for this session' % vardef_ctxt[0]
                    )
                    continue
                this_vals = _nested_get(
                    session_vals, vardef_ctxt
                )  # returns list of values for given session and context = column
                mean, std = np.mean(this_vals), np.std(this_vals)
                unit = _var_unit(vardef_ctxt)
                if unit == 'deg':
                    unit = u'\u00B0'  # Unicode degree sign
                else:
                    unit = ' ' + unit
                element += u'%s: %.2f±%.2f%s' % (ctxt, mean, std, unit)
            row.append(element)
        table.append(row)
    return _plot_tabular_data(table, row_labels, col_labels)


def _plot_extracted_table_plotly(curve_vals, vardefs):
    """Plot comparison of extracted gait curve values as a table, using Plotly as backend."""
    contexts = utils.get_contexts()
    # make a nested list of column headers; first row is session, second row is context
    session_labels = list(curve_vals.keys())
    col_labels_1 = list(
        itertools.chain.from_iterable(zip(session_labels, itertools.repeat('')))
    )
    context_cycle = itertools.cycle(c[0] for c in contexts)
    col_labels_2 = [next(context_cycle) for k in col_labels_1]
    # transpose
    col_labels = list(zip(col_labels_1, col_labels_2))
    row_labels = [_compose_varname(vardef) for vardef in vardefs]
    table = list()
    for vardef in vardefs:
        row = list()
        for session_vals in curve_vals.values():
            for ctxt, _ in contexts:
                vardef_ctxt = [ctxt + vardef[0]] + vardef[1:]
                if vardef_ctxt[0] not in session_vals:
                    logger.debug(
                        '%s was not collected for this session' % vardef_ctxt[0]
                    )
                    continue
                this_vals = _nested_get(
                    session_vals, vardef_ctxt
                )  # returns list of values for given session and context = column
                mean, std = np.mean(this_vals), np.std(this_vals)
                unit = _var_unit(vardef_ctxt)
                if unit == 'deg':
                    unit = u'\u00B0'  # Unicode degree sign
                else:
                    unit = ' ' + unit
                row.append(u'%.2f±%.2f%s' % (mean, std, unit))
        table.append(row)
    return _plot_tabular_data_via_plotly(table, row_labels, col_labels)


def _plot_tabular_data(data, row_labels=None, col_labels=None):
    """Plot tabular data via matplotlib table()"""
    fig = Figure()
    ax = fig.add_subplot(111)
    the_table = ax.table(
        cellText=data,
        rowLabels=row_labels,
        rowLoc='left',
        colLabels=col_labels,
        loc='center',
        bbox=[0.2, 0.0, 0.9, 0.8],
    )
    # the_table.auto_set_font_size(False)
    # the_table.set_fontsize(7)
    # the_table.scale(1, 1.5)
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    ax.set_frame_on(False)
    the_table.auto_set_font_size(False)
    the_table.set_fontsize(6)
    return fig


def _plot_tabular_data_via_plotly(data, row_labels=None, col_labels=None):
    """Plot tabular data via plotly and convert to matplotlib"""
    # disable MathJax since it causes slowdowns
    # see https://github.com/plotly/Kaleido/issues/36
    pio.kaleido.scope.mathjax = None
    # transpose into list of columns for go.Table
    data = list(zip(*data))
    data = [row_labels] + data
    col_labels = [''] + col_labels
    thetable = go.Table(cells={'values': data}, header={'values': col_labels})
    pfig = go.Figure(data=[thetable])
    bytes = pfig.to_image(format='png', engine='kaleido', width=1600, height=1200)
    bio = io.BytesIO(bytes)
    # 2: read into matplotlib
    im = matplotlib.image.imread(bio, format='png')
    fig = Figure()
    ax = fig.add_subplot(111)
    ax.imshow(im)
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    ax.set_frame_on(False)
    return fig


def _plot_vels(vels, labels, title=None):
    """Stem plot of trial velocities"""
    fig = Figure()
    ax = fig.add_subplot(111)
    ax.stem(vels)
    ax.set_xticks(range(len(vels)))
    ax.set_xticklabels(labels, rotation='vertical')
    ax.set_ylabel('Speed (m/s)')
    ax.tick_params(axis='both', which='major', labelsize=8)
    fig.set_tight_layout(True)
    if title:
        fig.suptitle(title)
    return fig


def _plot_timedep_vels(vels, labels, title=None):
    """Time-dependent velocity curves"""
    fig = Figure()
    ax = fig.add_subplot(111)
    ax.set_ylabel('Speed (m/s)')
    ax.set_xlabel('% of trial')
    for vel, label in zip(vels, labels):
        ax.plot(vel, label=label)
    fig.set_tight_layout(True)
    if title:
        fig.suptitle(title)
    # for legend, maybe wait for matplotlib 3
    # ax.legend()
    return fig


def _plot_height_ratios(layout):
    """Calculate height ratios for mpl plot"""
    plotheightratios = []
    for row in layout:
        if all([models.model_from_var(var) for var in row]):
            plotheightratios.append(1)
        else:
            plotheightratios.append(cfg.plot_matplotlib.analog_plotheight)
    return plotheightratios


def time_dist_barchart(
    values,
    stddev=None,
    thickness=0.5,
    color=None,
    stddev_bars=True,
    plotvars=None,
    figtitle=None,
    big_fonts=None,
):
    """Multi-variable and multi-condition barchart plot.

    Used mostly for plotting time-distance variables.

    Parameters
    ----------
    values : dict
        Nested dict of values to plot. Keys are values[condition][var][context].
        Matches the output of group_analysis_trials().
    stddev : dict | None
        Similar to values, but provides standard deviation for each variable.
    thickness : float, optional
        Y direction thickness of the bars, by default 0.5
    color : list, optional
        List of colors for the different conditions. If None, use default values.
    stddev_bars : bool, optional
        Whether to plot standard deviation also as bars.
    plotvars : list, optional
        The variables to plot and their order. If None, plot all variables.
    figtitle : str, optional
        Title of the plot.
    big_fonts : bool, optional
        If True, increase font sizes somewhat.

    Returns
    -------
    Figure
        The chart.
    """

    fig = Figure()

    def _plot_label(ax, rects, texts):
        """Plot a label inside each rect"""
        # FIXME: just a rough heuristic for font size
        fontsize = 6 if len(rects) >= 3 else 8
        for rect, txt in zip(rects, texts):
            ax.text(
                rect.get_width() * 0.0,
                rect.get_y() + rect.get_height() / 2.0,
                txt,
                ha='left',
                va='center',
                size=fontsize,
            )

    def _plot_oneside(vars, context, col):
        """Do the bar plots for given context and column"""
        for ind, var in enumerate(vars):
            ax = fig.add_subplot(gs[ind, col])
            if ind == 0:
                if col == 0:
                    ax.set_title('Left')
                elif col == 2:
                    ax.set_title('Right')

            ax.axis('off')
            # may have several bars (conditions) per variable
            vals_this = [values[cond][var][context] for cond in conds]
            if not np.count_nonzero(~np.isnan(vals_this)):
                continue
            if stddev is None:
                stddevs_this = [None for cond in conds]
            else:
                stddevs_this = [stddev[cond][var][context] for cond in conds]
            units_this = len(conds) * [units[ind]]
            ypos = np.arange(len(vals_this) * thickness, 0, -thickness)
            xerr = stddevs_this if stddev_bars else None
            rects = ax.barh(
                ypos, vals_this, thickness, align='edge', color=color, xerr=xerr
            )
            # FIXME: set axis scale according to var normal values
            ax.set_xlim([0, 1.5 * max(vals_this)])
            texts = list()
            for val, std, unit in zip(vals_this, stddevs_this, units_this):
                if val == 0:
                    texts += ['']
                elif std:
                    texts += [u'%.2f ± %.2f %s' % (val, std, unit)]
                else:
                    texts += [u'%.2f %s' % (val, unit)]
            _plot_label(ax, rects, texts)
        # return the last set of rects for legend
        return rects

    if color is None:
        color = [
            'tab:green',
            'tab:orange',
            'tab:red',
            'tab:brown',
            'tab:pink',
            'tab:gray',
            'tab:olive',
        ]

    conds, vars, units = _pick_common_vars(values, plotvars)

    # 3 columns: bars, labels, bars
    gs = gridspec.GridSpec(len(vars), 3, width_ratios=[1, 1 / 3.0, 1])

    # variable names into the center column
    for ind, (var, unit) in enumerate(zip(vars, units)):
        textax = fig.add_subplot(gs[ind, 1])
        textax.axis('off')
        label = '%s (%s)' % (var, unit) if unit else var
        textax.text(0, 0.5, label, ha='center', va='center')

    rects = _plot_oneside(vars, 'Left', 0)
    rects = _plot_oneside(vars, 'Right', 2)

    if len(conds) > 1:
        fig.legend(
            rects,
            conds,
            fontsize=7,
            bbox_to_anchor=(0.5, 0),
            loc="lower right",
            bbox_transform=fig.transFigure,
        )

    if figtitle is not None:
        fig.suptitle(figtitle)
    return fig


def _annotate_axis(ax, text):
    """Annotate at center of mpl axis"""
    ax.annotate(
        text,
        xy=(0.5, 0.5),
        xycoords='axes fraction',
        ha="center",
        va="center",
        fontsize=cfg.plot_matplotlib.subtitle_fontsize,
    )


def _remove_ticks_and_labels(ax):
    ax.tick_params(
        axis='both',
        which='both',
        bottom=False,
        top=False,
        labelbottom=False,
        right=False,
        left=False,
        labelleft=False,
    )


def plot_trials(
    trials,
    layout,
    model_normaldata=None,
    emg_normaldata=None,
    cycles=None,
    max_cycles=None,
    emg_mode=None,
    legend_type=None,
    style_by=None,
    color_by=None,
    supplementary_data=None,
    legend=True,
    figtitle=None,
):
    """Plot gait trials using matplotlib.

    Parameters
    ----------
    trials : list
        List of Trial instances to plot.
    layout : list
        The plot layout to use.
    model_normaldata : dict | None
        Normaldata for model variables. If None, taken from cfg.
    emg_normaldata : dict | None
        Normal data for EMG variables. If None, taken from cfg.
    cycles : dict | str | int | tuple | list
        Cycles to plot. See Trial.get_cycles() for details.
    max_cycles : dict | None
        Maximum number of cycles to plot for each variable type. If None, taken
        from cfg.
    emg_mode : str | None
        If 'rms', plot EMG in RMS mode.
    legend_type : str | None
        Legend type for gait cycles (see _get_cycle_name for options). If None,
        taken from cfg.
    style_by : dict | None
        How to style each variable type. If None, taken from cfg.
    color_by : dict | None
        How to color each variable type. If None, taken from cfg.
    supplementary_data : dict | None
        Supplementary data to plot for each variable.
    legend : bool
        If True, plot the legend.
    figtitle : str | None
        Main title for the figure.

    Returns
    -------
    Figure
        The matplotlib figure object.
    """
    if not trials:
        raise GaitDataError('No trials')

    if not isinstance(trials, list):
        trials = [trials]

    style_by, color_by = _handle_style_and_color_args(style_by, color_by)

    if legend_type is None:
        legend_type = 'short_name_with_cyclename'

    if supplementary_data is None:
        supplementary_data = dict()

    if model_normaldata is None:
        model_normaldata = normaldata._read_configured_model_normaldata()

    if emg_normaldata is None:
        emg_normaldata = normaldata._read_configured_emg_normaldata()

    use_rms = emg_mode == 'rms'

    nrows, ncols = layouts._check_layout(layout)

    # these generate and keep track of key -> linestyle (or color) mappings
    trace_colors = _cyclical_mapper(cfg.plot.colors)
    emg_trace_colors = _cyclical_mapper(cfg.plot.colors)
    trace_styles = _cyclical_mapper(cfg.plot.linestyles)

    # compute figure width and height
    figh = min(nrows * cfg.plot_matplotlib.inch_per_row + 1, cfg.plot_matplotlib.maxh)
    figw = min(ncols * cfg.plot_matplotlib.inch_per_col, cfg.plot_matplotlib.maxw)
    figw, figh = (20, 10)
    fig = Figure(figsize=(figw, figh), constrained_layout=True)

    plotheightratios = _plot_height_ratios(layout)
    plotheightratios.append(0.5)  # for legend
    gridspec_ = gridspec.GridSpec(
        nrows + 1, ncols, figure=fig, height_ratios=plotheightratios, width_ratios=None
    )
    # spacing adjustments for the plot, see Figure.tight_layout()
    # pad == plot margins
    # w_pad, h_pad == horizontal and vertical subplot padding
    # rect leaves extra space for figtitle
    # auto_spacing_params = dict(pad=.2, w_pad=.3, h_pad=.3,
    #                            rect=(0, 0, 1, .95))
    # fig = Figure(figsize=(figw, figh),
    #              tight_layout=auto_spacing_params)
    # fig = Figure(figsize=(figw, figh))

    normalized = cycles != 'unnormalized'
    cycles = _handle_cyclespec(cycles)
    if max_cycles is None:
        max_cycles = cfg.plot.max_cycles

    axes = dict()
    leg_entries = OrderedDict()
    mod_normal_lines_ = None
    emg_normal_lines_ = None
    emg_any_ok = defaultdict(lambda: False)

    # plot actual data
    for trial_ind, trial in enumerate(trials):

        cyclebunch = _get_trial_cycles(trial, cycles, max_cycles)
        # the idea here is to sort the trial cycles by their legend key, so they
        # appear in the legend in correct order
        sorter = partial(_get_cycle_name, trial, name_type=legend_type)
        allcycles = sorted(cyclebunch.allcycles, key=sorter)

        for cyc_ind, cyc in enumerate(allcycles):

            first_cyc = trial_ind == 0 and cyc_ind == 0
            context = cyc.context

            for i, row in enumerate(layout):
                if i not in axes:
                    axes[i] = dict()
                for j, var in enumerate(row):

                    # create axis or get existing axis to use
                    if j not in axes[i]:
                        sharex = axes[0][0] if i > 0 or j > 0 else None
                        ax = fig.add_subplot(gridspec_[i, j], sharex=sharex)
                        # set x axis to tightly match data boundaries
                        ax.autoscale(enable=True, axis='x', tight=True)
                        axes[i][j] = ax
                    else:
                        ax = axes[i][j]

                    vartype = _triage_var(var, trial)
                    if vartype is None:
                        ax.axis('off')
                        continue

                    # tracegroup is the legend entry for this cycle
                    # depending on name_type, one tracegroup may end up
                    # holding several cycles, which will be under the same
                    # legend entry.
                    cyclename = _get_cycle_name(trial, cyc, name_type=legend_type)
                    cyclename_full = _get_cycle_name(trial, cyc, name_type='full')

                    if vartype == 'model':
                        do_plot = cyc in cyclebunch.model_cycles
                        themodel = models.model_from_var(var)
                        if var in themodel.varnames_noside:
                            # var context was unspecified, so choose it
                            # according to cycle context
                            var = context + var
                        elif var[0] != context:
                            # specified var context does not match cycle
                            do_plot = False

                        # kinetic var cycles are required to have valid
                        # forceplate data
                        if (
                            normalized
                            and themodel.is_kinetic_var(var)
                            and not cyc.on_forceplate
                        ):
                            do_plot = False

                        # plot normal data before first cycle
                        if model_normaldata is not None and first_cyc and normalized:
                            nvar = var if var in themodel.varlabels_noside else var[1:]
                            key = nvar if nvar in model_normaldata else None
                            ndata = (
                                model_normaldata[key]
                                if key in model_normaldata
                                else None
                            )
                            if ndata is not None:
                                normalx = np.linspace(0, 100, ndata.shape[0])
                                mod_normal_lines_ = ax.fill_between(
                                    normalx,
                                    ndata[:, 0],
                                    ndata[:, 1],
                                    color=cfg.plot.model_normals_color,
                                    alpha=cfg.plot.model_normals_alpha,
                                )

                        t, y = trial.get_model_data(var, cycle=cyc)
                        if y is None:
                            do_plot = False

                        if do_plot:
                            # get style and color
                            sty = _style_by_params(
                                style_by['model'], trace_styles, trial, cyc, context
                            )
                            col = _color_by_params(
                                color_by['model'], trace_colors, trial, cyc, context
                            )

                            line_ = ax.plot(
                                t,
                                y,
                                color=col,
                                linestyle=sty,
                                linewidth=cfg.plot.model_linewidth,
                                alpha=cfg.plot.model_alpha,
                            )[0]
                            leg_entries[cyclename] = line_

                            # add toeoff marker
                            if cyc.toeoffn is not None:
                                toeoff_marker = ax.plot(
                                    t[cyc.toeoffn : cyc.toeoffn + 1],
                                    y[cyc.toeoffn : cyc.toeoffn + 1],
                                    col,
                                    marker='^',
                                )

                            # each cycle gets its own stddev plot
                            if isinstance(trial, AvgTrial):
                                model_stddev = trial.stddev_data
                                if (
                                    model_stddev is not None
                                    and normalized
                                    and y is not None
                                    and var in model_stddev
                                ):
                                    sdata = model_stddev[var]
                                    stdx = np.linspace(0, 100, sdata.shape[0])
                                    stddev_ = ax.fill_between(
                                        stdx,
                                        y - sdata,
                                        y + sdata,
                                        color=col,
                                        alpha=cfg.plot.model_stddev_alpha,
                                    )
                                    leg_entries['Stddev for %s' % cyclename] = stddev_

                            # add supplementary data
                            # XXX: not implemented for matplotlib

                            # axis adjustments for model variable
                            if not ax.get_ylabel():
                                yunit = themodel.units[var]
                                if yunit == 'deg':
                                    yunit = u'\u00B0'  # degree sign
                                ydesc = [s[:3] for s in themodel.ydesc[var]]  # shorten
                                ylabel_ = '%s %s %s' % (ydesc[0], yunit, ydesc[1])
                                ax.set(ylabel=ylabel_)

                                ax.xaxis.label.set_fontsize(
                                    cfg.plot_matplotlib.label_fontsize
                                )
                                ax.yaxis.label.set_fontsize(
                                    cfg.plot_matplotlib.label_fontsize
                                )
                                title = _var_title(var)
                                if title:
                                    ax.set_title(title)
                                    ax.title.set_fontsize(
                                        cfg.plot_matplotlib.subtitle_fontsize
                                    )
                                ax.tick_params(
                                    axis='both',
                                    which='major',
                                    labelsize=cfg.plot_matplotlib.ticks_fontsize,
                                )
                                ax.locator_params(axis='y', nbins=6)  # less tick marks

                                # FIXME: add n of averages for AvgTrial
                                is_avg_trial = False
                                if is_avg_trial:
                                    subplot_title += (
                                        ' (avg of %d cycles)' % trial.n_ok[var]
                                    )

                    # plot marker variable
                    elif vartype == 'marker':
                        do_plot = cyc in cyclebunch.model_cycles

                        t, mdata = trial.get_marker_data(var, cycle=cyc)
                        if mdata is None:
                            do_plot = False

                        if do_plot:

                            for datadim, data in zip('XYZ', mdata.T):
                                sty = _style_by_params(
                                    style_by['marker'],
                                    trace_styles,
                                    trial,
                                    cyc,
                                    context,
                                    datadim,
                                )
                                col = _color_by_params(
                                    color_by['marker'],
                                    trace_colors,
                                    trial,
                                    cyc,
                                    context,
                                    datadim,
                                )

                                line_ = ax.plot(
                                    t,
                                    data,
                                    color=col,
                                    linestyle=sty,
                                    linewidth=cfg.plot.model_linewidth,
                                    alpha=cfg.plot.model_alpha,
                                )[0]

                                # dim-specific tracename
                                tracename_marker = 'mkr_%s:%s' % (datadim, cyclename)
                                leg_entries[tracename_marker] = line_

                                # add toeoff marker
                                if cyc.toeoffn is not None:
                                    toeoff_marker = ax.plot(
                                        t[cyc.toeoffn : cyc.toeoffn + 1],
                                        data[cyc.toeoffn : cyc.toeoffn + 1],
                                        col,
                                        marker='^',
                                    )

                            # adjust subplot once
                            if (
                                not ax.get_ylabel()
                            ):  # this gets modified the first time around
                                ylabel = 'mm'
                                ax.set(ylabel=ylabel)

                                ax.xaxis.label.set_fontsize(
                                    cfg.plot_matplotlib.label_fontsize
                                )
                                ax.yaxis.label.set_fontsize(
                                    cfg.plot_matplotlib.label_fontsize
                                )
                                title = _var_title(var)
                                if title:
                                    ax.set_title(title)
                                    ax.title.set_fontsize(
                                        cfg.plot_matplotlib.subtitle_fontsize
                                    )
                                ax.tick_params(
                                    axis='both',
                                    which='major',
                                    labelsize=cfg.plot_matplotlib.ticks_fontsize,
                                )
                                ax.locator_params(axis='y', nbins=6)  # less tick marks

                    # plot EMG variable
                    elif vartype == 'emg':
                        do_plot = (
                            trial.emg is not None
                            and trial.emg.context_ok(var, context)
                            and trial.emg.status_ok(var)
                            and cyc in cyclebunch.emg_cycles
                        )
                        # FIXME: maybe annotate disconnected chans
                        # _no_ticks_or_labels(ax)
                        # _axis_annotate(ax, 'disconnected')
                        if do_plot:

                            t_, y = trial.get_emg_data(var, rms=use_rms, cycle=cyc)
                            t = t_ if normalized else t_ / trial.samplesperframe

                            col = _color_by_params(
                                color_by['emg'], emg_trace_colors, trial, cyc, context
                            )
                            lw = (
                                cfg.plot.emg_rms_linewidth
                                if use_rms
                                else cfg.plot.emg_linewidth
                            )
                            line_ = ax.plot(
                                t, y * cfg.plot.emg_multiplier, color=col, linewidth=lw
                            )[0]
                            leg_entries['EMG: ' + cyclename] = line_

                        # do normal data & plot adjustments for last EMG cycle
                        # keeps track of whether any trials have valid EMG data for this
                        # variable; otherwise, we annotatate channel as disconnected
                        emg_any_ok[var] |= trial.emg.status_ok(var)
                        remaining = allcycles[cyc_ind + 1 :]
                        last_emg = trial == trials[-1] and not any(
                            c in remaining for c in cyclebunch.emg_cycles
                        )
                        if last_emg:
                            title = _var_title(var)
                            if not emg_any_ok[var]:
                                _remove_ticks_and_labels(ax)
                                _annotate_axis(ax, '%s disconnected' % title)
                            else:
                                if title:
                                    ax.set_title(title)
                                    ax.title.set_fontsize(
                                        cfg.plot_matplotlib.subtitle_fontsize
                                    )
                                ax.set(ylabel=cfg.plot.emg_ylabel)
                                ax.yaxis.label.set_fontsize(
                                    cfg.plot_matplotlib.label_fontsize
                                )
                                ax.locator_params(axis='y', nbins=4)
                                # tick font size
                                ax.tick_params(
                                    axis='both',
                                    which='major',
                                    labelsize=cfg.plot_matplotlib.ticks_fontsize,
                                )

                                _emg_y_extent = _emg_yscale(emg_mode)
                                ax.set_ylim(_emg_y_extent)

                                if normalized and var in emg_normaldata:
                                    ndata = emg_normaldata[var][None, :]
                                    # create a color strip below the EMG trace, according to normal data
                                    extent_y0 = _emg_y_extent[0]
                                    # strip width is total y scale / 10
                                    extent_y1 = (
                                        extent_y0
                                        + (_emg_y_extent[1] - _emg_y_extent[0]) / 10.0
                                    )
                                    emg_normal_lines_ = ax.imshow(
                                        ndata,
                                        extent=[0, 100, extent_y0, extent_y1],
                                        aspect='auto',
                                        cmap='Reds',
                                        vmin=0,
                                        vmax=1,
                                    )

                    elif vartype == 'unknown':
                        raise GaitDataError('cannot interpret variable %s' % var)

                    else:
                        raise GaitDataError(
                            'plotting not implemented for variable %s' % var
                        )

                    # adjustments common to all plots
                    # set x labels on bottom row of plot
                    if row == layout[-1]:
                        xlabel = '% of gait cycle' if normalized else 'frame'
                        ax.set(xlabel=xlabel)
                        ax.xaxis.label.set_fontsize(cfg.plot_matplotlib.label_fontsize)

    if figtitle is not None:
        # constrained_layout does not work well with suptitle
        # (https://github.com/matplotlib/matplotlib/issues/13672)
        # hack: add extra \n to create whitespace
        # Py2: this hack is no longer necessary in matplotlib 3.x or newer
        if figtitle and figtitle[-1] != '\n':
            figtitle = figtitle + '\n'
        fig.suptitle(figtitle, fontsize=10)

    if legend:
        # put legend into its own axis, since constrained_layout does not handle fig.legend yet
        # see https://github.com/matplotlib/matplotlib/issues/13023
        axleg = fig.add_subplot(gridspec_[i + 1, :])
        axleg.axis('off')
        leg_entries_ = OrderedDict()
        if mod_normal_lines_:
            leg_entries_['Norm.'] = mod_normal_lines_
        if emg_normal_lines_:
            leg_entries_['EMG norm.'] = emg_normal_lines_
        leg_entries_.update(leg_entries)
        leg_ncols = ncols
        leg = axleg.legend(
            leg_entries_.values(),
            leg_entries_.keys(),
            fontsize=cfg.plot_matplotlib.legend_fontsize,
            loc='upper center',
            bbox_to_anchor=(0.5, 1.05),
            ncol=leg_ncols,
        )
        # legend lines may be too thin to see
        for li in leg.get_lines():
            li.set_linewidth(2.0)
    return fig



# -*- coding: utf-8 -*-
"""
C3D reader functions.

NB: do not use the data readers from this file directly. They are intended to be
called via the read_data module.

@author: Jussi (jnu@iki.fi)
"""

from __future__ import print_function, division
from builtins import zip
from past.builtins import basestring
from collections import defaultdict
import logging
import numpy as np
import os
import os.path as op
import sys
import shutil

from .numutils import center_of_pressure, _change_coords, _is_ascii
from .utils import TrialEvents, _step_width
from .envutils import lru_cache_checkfile, _named_tempfile, GaitDataError


logger = logging.getLogger(__name__)

# try the import the bundled btk version first (currently Python 2.7 64-bit
# only); if that fails, try other options
try:
    from .thirdparty import btk

    BTK_IMPORTED = True
except ImportError:
    try:
        import btk

        BTK_IMPORTED = True
    except ImportError:
        try:
            # try pyBTK package
            if sys.version_info.major == 3:
                from pyBTK.btk3 import btk
            elif sys.version_info.major == 2:
                from pyBTK.btk2 import btk
            else:
                raise Exception('unexpected major Python version')
            BTK_IMPORTED = True
        except ImportError:
            BTK_IMPORTED = False
            logger.warning('cannot import btk module; unable to read .c3d files')


def _is_c3d_file(source):
    """Check if source is a valid c3d file.

    XXX: Currently we just check existence.
    """
    return isinstance(source, basestring) and op.isfile(source)


def _get_c3d_metadata_subfields(acq, field):
    """Return names of metadata subfields for a given field"""
    meta = acq.GetMetaData()
    meta_s = meta.GetChild(field)
    return [f.GetLabel() for f in btk.Iterate(meta_s)]


def _get_c3d_metadata_field(acq, field, subfield):
    """Get c3d metadata FIELD:SUBFIELD as Python type.

    Always returns a list.
    """
    meta = acq.GetMetaData()

    def _get_child(field, child):
        try:
            return field.GetChild(child)
        except RuntimeError:
            raise RuntimeError('Invalid c3d metadata field: %s' % child)

    info = _get_child(_get_child(meta, field), subfield).GetInfo()
    if info.GetFormatAsString() == 'Char':
        return [s.strip() for s in info.ToString()]
    elif info.GetFormatAsString() == 'Real':
        return [x for x in info.ToDouble()]
    else:
        raise RuntimeError('Unhandled btk meta info type')


@lru_cache_checkfile
def _get_c3dacq(c3dfile):
    """Get a btk c3dacq object.

    Object is returned from cache if filename and digest match.
    """
    reader = btk.btkAcquisitionFileReader()
    # Py2: btk interface cannot take unicode, so encode to latin-1 first
    if sys.version_info.major == 2:
        c3dfile = c3dfile.encode('latin-1')
    reader.SetFilename(c3dfile)
    try:
        reader.Update()
    except RuntimeError as e:
        # this is a workaround for an unresolved BTK Python 3 bug, where
        # filenames containing extended characters cannot be read
        if not _is_ascii(c3dfile) and sys.version_info.major == 3:
            logger.warning('trying to work around possible btk extended chars bug')
            # just copy the file into another directory
            temp_path = _named_tempfile(suffix='.c3d')
            shutil.copy2(c3dfile, temp_path)
            logger.warning('using %s' % temp_path)
            # we need to create a new reader here
            # (the previous update call screws it up somehow)
            reader = btk.btkAcquisitionFileReader()
            reader.SetFilename(temp_path)
            reader.Update()
            # we should be able to remove the temp file now
            os.remove(temp_path)
        else:
            # something else went wrong during read
            raise e
    return reader.GetOutput()


def get_analysis(c3dfile, condition='unknown'):
    """Get analysis values from c3d (e.g. gait parameters).

    Parameters
    ----------
    c3dfile : str
        Name of the file.
    condition : str, optional
        The condition name, by default 'unknown'.

    Returns
    -------
    dict
        A nested dict of the analysis values, keyed by variable name and
        context. The first key is the condition name.
    """
    logger.debug('getting analysis values from %s' % c3dfile)
    acq = _get_c3dacq(c3dfile)
    try:
        vars_ = _get_c3d_metadata_field(acq, 'ANALYSIS', 'NAMES')
        units = _get_c3d_metadata_field(acq, 'ANALYSIS', 'UNITS')
        contexts = _get_c3d_metadata_field(acq, 'ANALYSIS', 'CONTEXTS')
        vals = _get_c3d_metadata_field(acq, 'ANALYSIS', 'VALUES')
    except RuntimeError:
        raise GaitDataError('Cannot read time-distance parameters from %s' % c3dfile)

    # build a nice output dict
    di = defaultdict(lambda: defaultdict(dict))
    di_ = di[condition]

    for (var, unit, context, val) in zip(vars_, units, contexts, vals):
        di_[var]['unit'] = unit
        di_[var][context] = val

    # if c3d was missing vals for some var/context, insert nans
    for var in di_:
        for context in ['Left', 'Right']:
            if context not in di_[var]:
                logger.warning(
                    '%s has missing value: %s / %s' % (c3dfile, var, context)
                )
                di_[var][context] = np.nan

    # Nexus <2.8 did not output step width into c3d, so compute it here if
    # needed
    if 'Step Width' not in di_:
        logger.warning('computing step widths (not found in %s)' % c3dfile)
        sw = _step_width(c3dfile)
        di[condition]['Step Width'] = dict()
        # XXX: currently uses average of all cycles from trial
        di[condition]['Step Width']['Right'] = np.array(sw['R']).mean()
        di[condition]['Step Width']['Left'] = np.array(sw['L']).mean()
        di[condition]['Step Width']['unit'] = 'm'
    return di


def _get_emg_data(c3dfile):
    """Read EMG data from a c3d file.

    See read_data.get_emg_data() for details.
    """
    return _get_analog_data(c3dfile, 'EMG')


def _get_accelerometer_data(c3dfile):
    """Read accelerometer data from a c3d file"""
    data = _get_analog_data(c3dfile, 'Accelerometer')
    # Remove the 'Acceleration.' prefix if inserted by Nexus, so that channel
    # names match Nexus. This is a bit ugly (not done for EMG which uses fuzzy
    # matching)
    for key in data['data']:
        if key.find('Acceleration.') == 0:
            # replace key names
            data['data'][key[13:]] = data['data'].pop(key)
    return data


def _get_analog_data(c3dfile, devname):
    """Read analog data from a c3d file.

    devname is matched against channel names.
    """
    acq = _get_c3dacq(c3dfile)
    data = dict()
    chnames = []
    for i in btk.Iterate(acq.GetAnalogs()):
        if i.GetDescription().find(devname) >= 0:
            chname = i.GetLabel()
            chnames.append(chname)
            data[chname] = np.squeeze(i.GetValues())
    if chnames:
        return {
            't': np.arange(len(data[chname])) / acq.GetAnalogFrequency(),
            'data': data,
        }
    else:
        raise GaitDataError('No matching analog channels found in data')


def _get_marker_data(c3dfile, markers, ignore_missing=False):
    """Get position data for specified markers.

    See read_data.get_marker_data for details.
    """
    if not isinstance(markers, list):  # listify if not already a list
        markers = [markers]
    acq = _get_c3dacq(c3dfile)
    mkrdata = dict()
    for marker in markers:
        try:
            mkrdata[marker] = np.squeeze(acq.GetPoint(marker).GetValues())
        except RuntimeError:
            if ignore_missing:
                logger.warning('Cannot read trajectory %s from c3d file' % marker)
                continue
            else:
                raise GaitDataError('Cannot read trajectory %s from c3d file' % marker)
    return mkrdata


def _get_c3d_subject_param(acq, param):
    try:
        param = _get_c3d_metadata_field(acq, 'PROCESSING', param)[0]
    except RuntimeError:
        logger.warning('Cannot get subject parameter %s' % param)
        param = None
    return param


def _get_metadata(c3dfile):
    """Read trial and subject metadata from c3d file.

    See read_data.get_metadata() for details.
    """
    trialname = os.path.basename(os.path.splitext(c3dfile)[0])
    sessionpath = os.path.dirname(c3dfile)
    acq = _get_c3dacq(c3dfile)
    # frame offset (start of trial data in frames)
    offset = acq.GetFirstFrame()
    lastfr = acq.GetLastFrame()
    length = lastfr - offset + 1  # or acq.GetPointFrameNumber()
    framerate = acq.GetPointFrequency()
    analograte = acq.GetAnalogFrequency()
    samplesperframe = acq.GetNumberAnalogSamplePerFrame()

    # count forceplates
    fpe = btk.btkForcePlatformsExtractor()
    fpe.SetInput(acq)
    fpe.Update()
    n_forceplates = len(list(btk.Iterate(fpe.GetOutput())))

    # get markers
    try:
        markers = _get_c3d_metadata_field(acq, 'POINT', 'LABELS')
    except RuntimeError:
        markers = list()
    # not sure what the '*xx' markers are, but delete them for now
    markers = [m for m in markers if m[0] != '*']

    #  get events
    rstrikes, lstrikes, rtoeoffs, ltoeoffs = [], [], [], []
    for i in btk.Iterate(acq.GetEvents()):
        if i.GetLabel() == "Foot Strike":
            if i.GetContext() == "Right":
                rstrikes.append(i.GetFrame())
            elif i.GetContext() == "Left":
                lstrikes.append(i.GetFrame())
            else:
                raise GaitDataError("Unknown context on foot strike event")
        elif i.GetLabel() == "Foot Off":
            if i.GetContext() == "Right":
                rtoeoffs.append(i.GetFrame())
            elif i.GetContext() == "Left":
                ltoeoffs.append(i.GetFrame())
            else:
                raise GaitDataError("Unknown context on foot strike event")
    events = TrialEvents(
        rstrikes=rstrikes, lstrikes=lstrikes, rtoeoffs=rtoeoffs, ltoeoffs=ltoeoffs
    )

    # get subject info
    try:
        name = _get_c3d_metadata_field(acq, 'SUBJECTS', 'NAMES')[0]
    except RuntimeError:
        logger.warning('Cannot get subject name')
        name = u'Unknown'

    try:
        par_names = _get_c3d_metadata_subfields(acq, 'PROCESSING')
    except RuntimeError:
        raise GaitDataError('%s is missing required subject info' % c3dfile)
    subj_params = defaultdict(lambda: None)
    subj_params.update({par: _get_c3d_subject_param(acq, par) for par in par_names})

    return {
        'trialname': trialname,
        'sessionpath': sessionpath,
        'offset': offset,
        'framerate': framerate,
        'analograte': analograte,
        'name': name,
        'subj_params': subj_params,
        'events': events,
        'length': length,
        'samplesperframe': samplesperframe,
        'n_forceplates': n_forceplates,
        'markers': markers,
    }


def _get_model_data(c3dfile, model):
    """Read model output variables (e.g. Plug-in Gait).

    See read_data.get_model_data for details.
    """
    modeldata = dict()
    acq = _get_c3dacq(c3dfile)
    var_dims = (3, acq.GetPointFrameNumber())
    for var in model.read_vars:
        try:
            vals = acq.GetPoint(var).GetValues()
            modeldata[var] = np.transpose(np.squeeze(vals))
        except RuntimeError:
            logger.info('cannot read model variable %s, returning nans' % var)
            data = np.empty(var_dims)
            data[:] = np.nan
            modeldata[var] = data
        # c3d stores scalars as last dim of 3-d array
        if model.read_strategy == 'last':
            modeldata[var] = modeldata[var][2, :]
    return modeldata


def _get_forceplate_data(c3dfile):
    """Read data of all forceplates from c3d file.

    See read_data.get_forceplate_data() for details.
    """
    logger.debug('reading forceplate data from %s' % c3dfile)
    read_chs = ['Fx', 'Fy', 'Fz', 'Mx', 'My', 'Mz']
    acq = _get_c3dacq(c3dfile)
    fpe = btk.btkForcePlatformsExtractor()
    fpe.SetInput(acq)
    fpe.Update()
    fpdata = list()
    nplate = 0
    for plate in btk.Iterate(fpe.GetOutput()):
        logger.debug('reading from plate %d' % nplate)
        nplate += 1
        if plate.GetType() != 2:
            # Nexus should always write forceplates as type 2
            raise GaitDataError('Only type 2 forceplates are supported for now')
        rawdata = dict()
        data = dict()
        for ch in btk.Iterate(plate.GetChannels()):
            label = ch.GetLabel()[-3:-1]  # strip descriptor and plate number
            rawdata[label] = np.squeeze(ch.GetData().GetValues())
        if not all([ch in rawdata for ch in read_chs]):
            logger.warning('could not read force/moment data for plate %d' % nplate)
            continue
        F = np.stack([rawdata['Fx'], rawdata['Fy'], rawdata['Fz']], axis=1)
        M = np.stack([rawdata['Mx'], rawdata['My'], rawdata['Mz']], axis=1)
        # we need to calculate center of pressure, since it's not in the c3d
        # this should be the plate thickness (from moment origin to physical
        # origin) needed for center of pressure calculations
        dz = np.abs(plate.GetOrigin()[2])
        cop = center_of_pressure(F, M, dz)  # in plate local coords
        Ftot = np.linalg.norm(F, axis=1)
        # locations of +x+y, -x+y, -x-y, +x-y plate corners in world coords
        # (in that order)
        cor = plate.GetCorners()
        wT = np.mean(cor, axis=1)  # translation vector, plate -> world
        # upper and lower bounds of forceplate
        ub = np.max(cor, axis=1)
        lb = np.min(cor, axis=1)
        # plate unit vectors in world system
        px = cor[:, 0] - cor[:, 1]
        py = cor[:, 0] - cor[:, 3]
        pz = np.array([0, 0, -1])
        P = np.stack([px, py, pz], axis=1)
        wR = P / np.linalg.norm(P, axis=0)  # rotation matrix, plate -> world
        # check whether CoP stays inside forceplate area and clip if necessary
        cop_w = _change_coords(cop, wR, wT)
        cop_wx = np.clip(cop_w[:, 0], lb[0], ub[0])
        cop_wy = np.clip(cop_w[:, 1], lb[1], ub[1])
        if not (cop_wx == cop_w[:, 0]).all() and (cop_wy == cop_w[:, 1]).all():
            logger.warning(
                'center of pressure outside forceplate bounds, clipping to plate'
            )
            cop[:, 0] = cop_wx
            cop[:, 1] = cop_wy
        # XXX moment and force transformations may still be wrong
        data['F'] = _change_coords(-F, wR, 0)  # not sure why sign flip needed
        data['Ftot'] = Ftot
        data['M'] = _change_coords(-M, wR, 0)  # not sure why sign flip needed
        data['CoP'] = cop_w
        data['wR'] = wR
        data['wT'] = wT
        data['plate_corners'] = cor.T
        fpdata.append(data)
    return fpdata
